from token import tok_name, NAME, NUMBER, STRING, ISNONTERMINAL, EQUAL
from symbol import sym_name
from new import instancemethod
import token, symbol, parser, sys

__all__ = [
    'parse_expr', 'build'
]

_name   = lambda builder,nodelist: builder.Name(nodelist[1])
_const  = lambda builder,nodelist: builder.Const(eval(nodelist[1]))


production = {
    NAME:   _name,
    NUMBER: _const,
    STRING: _const,
}


ops = {
    token.LEFTSHIFT: 'LeftShift',
    token.RIGHTSHIFT: 'RightShift',
    token.PLUS: 'Add',
    token.MINUS: 'Sub',
    token.STAR: 'Mul',
    token.SLASH: 'Div',
    token.PERCENT: 'Mod',
    token.DOUBLESLASH: 'FloorDiv',
}

def left_assoc(builder, nodelist):
    return getattr(builder,ops[nodelist[-2][0]])(nodelist[:-2],nodelist[-1])








def curry(f,*args):
    for arg in args:
        f = instancemethod(f,arg,type(arg))
    return f


def com_binary(opname, builder,nodelist):
    "Compile 'NODE (OP NODE)*' into (type, [ node1, ..., nodeN ])."
    items = [nodelist[i] for i in range(1,len(nodelist),2)]
    return getattr(builder,opname)(items)

# testlist: test (',' test)* [',']
# exprlist: expr (',' expr)* [',']
# subscriptlist: subscript (',' subscript)* [',']
testlist = testlist1 = exprlist = subscriptlist = curry(com_binary, 'Tuple')

# test: and_test ('or' and_test)* | lambdef
test = curry(com_binary, 'Or')

if sys.version>='2.5':
    or_test = test

    # test: or_test ['if' or_test 'else' test] | lambdef
    def test(builder, nodelist):
        return builder.IfElse(nodelist[1], nodelist[3], nodelist[5])

# and_test: not_test ('and' not_test)*
and_test = curry(com_binary, 'And')


# not_test: 'not' not_test | comparison
def not_test(builder, nodelist):
    return builder.Not(nodelist[2])








# comparison: expr (comp_op expr)*
def comparison(builder, nodelist):

    if len(nodelist)>4 and builder.simplify_comparisons:
        # Reduce (x < y < z ...) to (x<y and y<z and ...)
        return builder.And(
            [nodelist[:1]+nodelist[i:i+3] for i in range(1,len(nodelist)-1,2)]
        )

    results = []
    for i in range(3, len(nodelist), 2):
        nl = nodelist[i-1]

        # comp_op: '<' | '>' | '>=' | '<=' | '<>' | '!=' | '=='
        #          | 'in' | 'not' 'in' | 'is' | 'is' 'not'
        n = nl[1]
        if n[0] == token.NAME:
            type = n[1]
            if len(nl) == 3:
                if type == 'not':
                    type = 'not in'
                else:
                    type = 'is not'
        else:
            type = n[1]

        results.append((type, nodelist[i]))

    return builder.Compare(nodelist[1], results)


# expr: xor_expr ('|' xor_expr)*
expr = curry(com_binary, 'Bitor')

# xor_expr: and_expr ('^' and_expr)*
xor_expr = curry(com_binary, 'Bitxor')

# and_expr: shift_expr ('&' shift_expr)*
and_expr = curry(com_binary, 'Bitand')


# shift_expr: arith_expr ('<<'|'>>' arith_expr)*
# arith_expr: term (('+'|'-') term)*
# term: factor (('*'|'/'|'%'|'//') factor)*
shift_expr = arith_expr = term = left_assoc


unary_ops = {
    token.PLUS: 'UnaryPlus', token.MINUS: 'UnaryMinus', token.TILDE: 'Invert',
}


# factor: ('+'|'-'|'~') factor | power
def factor(builder, nodelist):
    return getattr(builder,unary_ops[nodelist[1][0]])(nodelist[2])



























# power: atom trailer* ['**' factor]

def power(builder, nodelist):
    if nodelist[-2][0]==token.DOUBLESTAR:
        return builder.Power(nodelist[:-2], nodelist[-1])

    node = nodelist[-1]
    nodelist = nodelist[:-1]
    t = node[1][0]

    # trailer: '(' [arglist] ')' | '[' subscriptlist ']' | '.' NAME

    if t == token.LPAR:
        return com_call_function(builder,nodelist,node[2])

    elif t == token.DOT:
        return builder.Getattr(nodelist, node[2][1])

    elif t == token.LSQB:
        item = node[2]

        while len(item)==2:
            item = item[1]

        if item[0]==token.COLON:
            lineno = item[2]
            return builder.Subscript(
                nodelist,
                (symbol.subscript, item, None)  # XXX optimization bypass
            )

        return builder.Subscript(nodelist, item)

    raise AssertionError("Unknown power", nodelist)







# atom: '(' [yield_expr|testlist_gexp] ')' |
#       '[' [listmaker] ']' |
#       '{' [dictmaker] '}' |
#       '`' testlist1 '`' |
#       NAME | NUMBER | STRING+

def atom(builder, nodelist):
    t = nodelist[1][0]
    if t == token.LPAR:
        if nodelist[2][0] == token.RPAR:
            return builder.Tuple(())
        return build(builder,nodelist[2])
    elif t==token.LSQB:
        if nodelist[2][0] == token.RSQB:
            return builder.List(())
        return listmaker(builder,nodelist[2])
    elif t==token.LBRACE:
        if nodelist[2][0] == token.RBRACE:
            items = ()
        else:
            dm = nodelist[2]
            items = [(dm[i],dm[i+2]) for i in range(1,len(dm),4)]
        return builder.Dict(items)
    elif t==token.BACKQUOTE:
        return builder.Backquote(nodelist[2])
    elif t==token.STRING:
        return builder.Const(eval(' '.join([n[1] for n in nodelist[1:]])))

    raise AssertionError("Unknown atom", nodelist)












# arglist: (argument ',')* (argument [',']| '*' test [',' '**' test] | '**' test)

def com_call_function(builder, primaryNode, nodelist):
    if nodelist[0] == token.RPAR:
        return builder.CallFunc(primaryNode,(),(),None,None)

    args = []; kw = []
    len_nodelist = len(nodelist)
    for i in range(1, len_nodelist, 2):
        node = nodelist[i]
        if node[0] == token.STAR or node[0] == token.DOUBLESTAR:
            break
        iskw, result = com_argument(node, kw)
        if iskw:
            kw.append(result)
        else:
            args.append(result)
    else:
        # No broken by star arg, so skip the last one we processed.
        i = i + 1

    if i < len_nodelist and nodelist[i][0] == token.COMMA:
        # need to accept an application that looks like "f(a, b,)"
        i = i + 1

    star_node = dstar_node = None
    while i < len_nodelist:
        tok = nodelist[i]
        ch = nodelist[i+1]
        i = i + 3
        if tok[0]==token.STAR:
            star_node = ch
        elif tok[0]==token.DOUBLESTAR:
            dstar_node = ch
        else:
            raise AssertionError, 'unknown node type: %s' % (tok,)

    return builder.CallFunc(primaryNode, args, kw, star_node, dstar_node)



# argument: [test '='] test [gen_for]  (Really [keyword '='] test)
# argument: test [gen_for] | test '=' test  # Really [keyword '='] test

def com_argument(nodelist, kw):
    if len(nodelist) == 2:
        if kw:
            raise SyntaxError, "non-keyword arg after keyword arg"
        return 0, nodelist[1]

    if nodelist[2][0] != token.EQUAL and len(nodelist)==3:
        return 0, (testlist_gexp.symbol, nodelist[1], nodelist[2])
    elif len(nodelist) !=4:
        raise AssertionError

    n = nodelist[1]
    while len(n) == 2 and n[0] != token.NAME:
        n = n[1]
    if n[0] != token.NAME:
        raise SyntaxError, "keyword can't be an expression (%r)" % (n,)

    return 1, ((token.STRING,`n[1]`,n[2]), nodelist[3])

# listmaker: test ( list_for | (',' test)* [','] )

def listmaker(builder, nodelist):

    values = []

    for i in range(1, len(nodelist)):

        if nodelist[i][0] == symbol.list_for:
            assert i==2 and len(nodelist)==3 and len(values)==1
            return com_iterator(builder.ListComp,values[0],nodelist[i])

        if nodelist[i][0] == token.COMMA:
            continue
        values.append(nodelist[i])

    return builder.List(values)


# list_iter: list_for | list_if
# list_for: 'for' exprlist 'in' testlist_safe [list_iter]
# list_if: 'if' old_test [list_iter]
#
# gen_iter: gen_for | gen_if
# gen_for: 'for' exprlist 'in' or_test [gen_iter]
# gen_if: 'if' old_test [gen_iter]

def com_iterator(method, value, nodelist):
    clauses = []
    nodelist = nodelist[1:]     # skip the symbol
    while nodelist:
        while len(nodelist)==1:
            nodelist, = nodelist
            nodelist = nodelist[1:]     # skip the symbol
        tok, val, line = nodelist[0]
        assert tok==NAME
        assert val in ('for', 'in', 'if')
        clauses.append((val, nodelist[1]))
        nodelist = nodelist[2:]
    return method(value, clauses)


# testlist_gexp test (gen_for | (',' test)* [','])

def testlist_gexp(builder, nodelist):
    if nodelist[2][0] == token.COMMA:
        return testlist(builder, nodelist)
    else:
        value, nodelist = nodelist[1:]
        return com_iterator(builder.GenExpr, value, nodelist)
    
testlist_comp = testlist_gexp   # Python 2.7

if hasattr(symbol, 'testlist_comp'):
    testlist_comp.symbol = symbol.testlist_comp
else:
    testlist_gexp.symbol = getattr(symbol, 'testlist_gexp', None)



# subscript: '.' '.' '.' | test | [test] ':' [test] [sliceop]
# sliceop: ':' [test]

def subscript(builder, nodelist):
    if nodelist[1][0]==token.DOT:
        return builder.Const(Ellipsis)

    item = nodelist; nl = len(nodelist)
    while type(item[1]) is tuple: item=item[1]   # find token, to get a line#
    lineno = item[-1]
    have_stride = nodelist[-1] and nodelist[-1][0]==symbol.sliceop
    if have_stride:
        start = stop = stride = None#; stride = (token.STRING, 'None', lineno)
        if len(nodelist[-1])==3:
            stride = nodelist[-1][2]
    else:
        start = None #(token.NUMBER,`0`,lineno)
        stop  = stride = None #(token.NUMBER,`sys.maxint`,lineno)

    if nl==5:
        start,stop = nodelist[1],nodelist[3]        # test : test sliceop
    elif nl==4:
        if nodelist[1][0]==token.COLON:             #   : test sliceop
            stop = nodelist[2]
        elif have_stride:                           # test :   sliceop
            start = nodelist[1]
        else:
            start, stop = nodelist[1], nodelist[3]  # test : test
    elif nl==3:
        if nodelist[1][0]==token.COLON:
            if not have_stride:
                stop = nodelist[2]      # : test
        else:
            start = nodelist[1]         # test :
    else:
        raise AssertionError("Unrecognized subscript", nodelist)
    if have_stride:
        return builder.Slice3(start,stop,stride)
    return builder.Slice2(start,stop)


for sym,name in sym_name.items():
    if name in globals():
        production[sym] = globals()[name]


def build(builder, nodelist):
    while len(nodelist)==2:
        nodelist = nodelist[1]
    return production[nodelist[0]](builder,nodelist)


def parse_expr(expr,builder):
    # include line numbers in parse data so valid symbols are never of length 2
    return build(builder, parser.expr(expr).totuple(1)[1])






























from peak.util.assembler import *
from core import *
from core import class_or_type_of, call_thru, flatten
from criteria import *
from indexing import *
from codegen import SMIGenerator, ExprBuilder, Getitem, IfElse, Tuple
from peak.util.decorators import decorate, synchronized, decorate_assignment
from types import InstanceType, ClassType
from ast_builder import build, parse_expr
import inspect, new, codegen, parser

__all__ = [
    'IsInstance', 'IsSubclass', 'Truth', 'Identity', 'Comparison', 'priority',
    'IndexedEngine', 'predicate_node_for', 'meta_function', 'expressionSignature',
]

abstract()
def predicate_node_for(builder, expr, cases, remaining_exprs, memo):
    """Return a dispatch tree node argument appropriate for the expr"""

def value_check(val, (exact, ranges)):
    if val in exact:
        return exact[val]
    lo = 0
    hi = len(ranges)
    while lo<hi:
        mid = (lo+hi)//2
        (tl,th), node = ranges[mid]
        if val<tl:
            hi = mid
        elif val>th:
            lo = mid+1
        else:
            return node
    raise AssertionError("Should never get here")

nodetype()
def IsInstance(expr, code=None):
    if code is None: return expr,
    return IsSubclass(class_or_type_of(expr), code)

_unpack = lambda c: c.UNPACK_SEQUENCE(2)
subclass_check = TryExcept(
    Suite([
        Code.DUP_TOP, SMIGenerator.ARG, _unpack, Code.ROT_THREE,
        Code.POP_TOP, Code.BINARY_SUBSCR, Code.ROT_TWO, Code.POP_TOP
    ]), [(Const(KeyError), Suite([
        SMIGenerator.ARG, _unpack, Code.POP_TOP, Call(Code.ROT_TWO, (Pass,)),
    ]))]
)

nodetype()
def IsSubclass(expr, code=None):
    if code is None: return expr,
    code(expr, subclass_check)

identity_check = IfElse(
    Getitem(SMIGenerator.ARG, Code.ROT_TWO),
    Compare(Code.DUP_TOP, [('in', SMIGenerator.ARG)]),
    Suite([Code.POP_TOP, Getitem(SMIGenerator.ARG, None)])
)

nodetype()
def Identity(expr, code=None):
    if code is None: return expr,
    code(Call(Const(id), (expr,), fold=False), identity_check)

nodetype()
def Comparison(expr, code=None):
    if code is None: return expr,
    code.LOAD_CONST(value_check)
    Call(Pass, (expr, SMIGenerator.ARG), code=code)

nodetype()
def Truth(expr, code=None):
    if code is None: return expr,
    skip = Label()
    code(SMIGenerator.ARG); code.UNPACK_SEQUENCE(2)
    code(expr, skip.JUMP_IF_TRUE, Code.ROT_THREE, skip, Code.POP_TOP,
         Code.ROT_TWO, Code.POP_TOP)


class priority(int):
    """An integer priority for manually resolving a rule ambiguity"""
when(implies, (priority, priority))(lambda p1,p2: p1>=p2)

class ExprBuilder(ExprBuilder):
    """Extended expression builder with support for meta-functions"""

    def Backquote(self, expr):
        raise SyntaxError("backquotes are not allowed in predicates")

meta_functions = {}

def meta_function(*stub, **parsers):
    """Declare a meta-function and its argument parsers"""
    stub, = stub
    def callback(frame, name, func, old_locals):
        for name in inspect.getargs(func.func_code)[0]:
            if not isinstance(name, basestring):
                raise TypeError(
                    "Meta-functions cannot have packed-tuple arguments"
                )
        what = func, parsers, inspect.getargspec(func)
        meta_functions[stub] = (
            lambda builder, *args: apply_meta(builder, what, *args)
        )
        return func
    return decorate_assignment(callback)

def expressionSignature(expr):
    """Convert raw Python code into logical conditions"""
    # Default is to simply test the truth of the expression
    return Test(Truth(expr), Value(True))

when(expressionSignature, (Const,))(lambda expr: bool(expr.value))

when(expressionSignature, ((bool, Test, Signature, Disjunction),))
def pass_through(expr):
    return expr



class CriteriaBuilder(ExprBuilder):
    simplify_comparisons = True

    def build_with(self, expr):
        self.push()
        try:
            return build(self, expr)
        finally:
            self.pop()

    def Not(self, expr):
        return codegen.Not(self.build_with(expr))

    def Or(self, items):
        return codegen.Or(map(self.build_with, items))

    def CallFunc(self, func, args, kw, star_node, dstar_node):
        b = build.__get__(self)
        target = b(func)
        if isinstance(target, Const) and target.value in meta_functions:
            return meta_functions[target.value](
                self, args, kw, star_node, dstar_node
            )
        return Call(
            target, map(b,args), [(b(k),b(v)) for k,v in kw],
            star_node and b(star_node), dstar_node and b(dstar_node)
        )

    def parse(self, expr):
        return expressionSignature(ExprBuilder.parse(self, expr))


when(expressionSignature, codegen.And)
def do_intersect(expr):
    return reduce(intersect, map(expressionSignature, expr.values), True)

when(expressionSignature, codegen.Or)
def do_union(expr):
    return OrElse(map(expressionSignature, expr.values))


when(expressionSignature, codegen.Not)
def do_negate(expr):
    return negate(expressionSignature(expr.expr))

_mirror_ops = {
    '>': '<', '>=': '<=', '=>':'<=',
    '<': '>', '<=': '>=', '=<':'>=',
    '<>': '<>', '!=': '<>', '==':'==',
    'is': 'is', 'is not': 'is not'
}

when(expressionSignature, codegen.Compare)
def do_compare(expr):
    left = expr.expr
    (op, right), = expr.ops

    if isinstance(left, Const) and op in _mirror_ops:
        left, right, op = right, left, _mirror_ops[op]

    if isinstance(right, Const):
        if op=='in' or op=='not in':
            cond = compileIn(left, right.value)
            if cond is not None:
                return maybe_invert(cond, op=='in')
        elif op=='is' or op=='is not':
            return maybe_invert(compileIs(left, right.value), op=='is')
        else:
            return Test(Comparison(left), Inequality(op, right.value))

    # Both sides involve variables or an un-optimizable constant,
    #  so it's a generic boolean criterion  :(
    return Test(Truth(expr), Value(True))









def apply_meta(builder,
    (func, parsers, (argnames, varargs, varkw, defaults)), args, kw, star, dstar
):
    # NB: tuple-args not allowed!
    def parse(arg, node):
        if not node:
            return None
        return parsers.get(arg, build)(builder, node)

    data = {}
    extra = []
    offset = 0
    for name in argnames:
        if name=='__builder__': data[name] = builder
        elif name=='__star__':  data[name] = parse(name, star)
        elif name=='__dstar__': data[name] = parse(name, dstar)
        else:
            break
        offset += 1

    for k, v in zip(argnames[offset:], args):
        data[k] = parse(k, v)

    varargpos = len(argnames)-offset
    if len(args)> varargpos:
        if not varargs:
            raise TypeError("Too many arguments for %r" % (func,))
        extra.extend([parse(varargs, node) for node in args[varargpos:]])

    for k,v in kw:
        k = build(builder, k)
        assert type(k) is Const and isinstance(k.value, basestring)
        k = k.value
        if k in data:
            raise TypeError("Duplicate keyword %s for %r" % (k,func))

        if varkw and k not in argnames and k not in parsers:
            data[k] = parse(varkw,  v)
        else:
            data[k] = parse(k, v)

    if star and '__star__' not in data:
        raise TypeError("%r does not support parsing *args" % (func,))

    if dstar and '__dstar__' not in data:
        raise TypeError("%r does not support parsing **kw" % (func,))

    if defaults:
        for k,v in zip(argnames[-len(defaults):], defaults):
            data.setdefault(k, v)

    try:
        args = map(data.pop, argnames)+extra
    except KeyError, e:
        raise TypeError(
            "Missing positional argument %s for %r"%(e.args[0], func)
        )
    return func(*args, **data)


def compile_let(builder, args, kw, star, dstar):
    """Compile the let() function"""
    if args or star or dstar:
        raise TypeError("let() only accepts inline keyword arguments")

    for k,v in kw:
        k = build(builder, k)
        assert type(k) is Const and isinstance(k.value, basestring)
        k = k.value
        v = build(builder, v)
        builder.bind({k:v})
    return True

from peak.rules import let
meta_functions[let] = compile_let







def _expand_as(func, predicate_string, *namespaces):
    """Pre-parse predicate string and register meta function"""

    args, varargs, kw, defaults = arginfo = inspect.getargspec(func)
    argnames = list(flatten(filter(None, [args, varargs, kw])))
    parsed = parser.expr(predicate_string).totuple(1)[1]
    builder = CriteriaBuilder(
        dict([(arg,Local(arg)) for arg in argnames]), *namespaces
    )
    bindings = {}
    for b in builder.bindings[-len(namespaces):][::-1]:
        bindings.update(b)

    # Make a function that just gets the arguments we want
    c = Code.from_function(func)
    c.return_(Call(Const(locals),fold=False))
    getargs = new.function(
        c.code(), func.func_globals, func.func_name, func.func_defaults,
        func.func_closure
    )

    def expand(builder, *args):        
        builder.push(bindings)   # globals, locals, etc.
        builder.bind(apply_meta(builder, (getargs, {}, arginfo), *args))

        # build in the newly-isolated namespace
        result = build(builder, parsed) 
        builder.pop()
        return result

    meta_functions[func] = expand

    c = Code.from_function(func)
    c.return_()
    if func.func_code.co_code == c.code().co_code:  # function body is empty
        c = Code.from_function(func)
        c.return_(build(builder, parsed))
        func.func_code = c.code()

    return func

def compileIs(expr, criterion):
    """Return a signature or predicate for 'expr is criterion'"""
    #if criterion is None:     # XXX this should be smarter
    #    return Test(IsInstance(expr), istype(NoneType))
    #else:
    return Test(Identity(expr), IsObject(criterion))

def maybe_invert(cond, truth):
    if not truth: return negate(cond)
    return cond

def compileIn(expr, criterion):
    """Return a signature or predicate (or None) for 'expr in criterion'"""
    try:
        iter(criterion)
    except TypeError:
        pass    # treat the in condition as a truth expression
    else:
        expr = Comparison(expr)
        return Test(expr, Disjunction([Value(v) for v in criterion]))

when(compileIn, (object, (type, ClassType)))
def compileInClass(expr, criterion):
    warn_parse("'x in SomeClass' syntax is deprecated; use 'isinstance(x,SomeClass)'")
    return Test(IsInstance(expr), Class(criterion))

when(compileIn, (object, istype))
def compileInIsType(expr, criterion):
    warn_parse("'x in istype(y)' syntax is deprecated; use 'type(x) is y'")
    return Test(IsInstance(expr), criterion)











def warn_parse(message, category=DeprecationWarning):
    """Issue a warning about a parsed string"""
    from warnings import warn, warn_explicit

    # Find the original call to _parse_string() to get its ParseContext
    import sys
    frame = sys._getframe(3)
    code = _parse_string.func_code
    ct = 2
    while frame is not None and frame.f_code is not code:
        frame = frame.f_back
        ct += 1
    if frame is None:
        # XXX direct use of expressionSignature; can't pinpoint a location
        return warn(message, category, ct)
    ctx = frame.f_locals['ctx']

    # Issue a warning against the method body
    g = ctx.globaldict
    #lineno = getattr(getattr(ctx.body, 'func_code', None), 'co_firstlineno', 2)
    module = g.get('__name__', "<string>")
    filename = g.get('__file__')
    if filename:
        fnl = filename.lower()
        if fnl.endswith(".pyc") or fnl.endswith(".pyo"):
            filename = filename[:-1]
    else:
        if module == "__main__":
            filename = sys.argv[0]
        if not filename:
            filename = module

    return warn_explicit(
        message, category, filename, ctx.lineno,
        g.setdefault("__warningregistry__", {})
    )





class IndexedEngine(Engine, TreeBuilder):
    """A dispatching engine that builds trees using bitmap indexes"""

    def __init__(self, disp):
        self.signatures = []
        self.all_exprs = {}
        super(IndexedEngine, self).__init__(disp)
        self.arguments = dict([(arg,Local(arg)) for arg in self.argnames])

    def _add_method(self, signature, rule):
        signature = Signature(tests_for(signature, self))
        if signature not in self.registry:
            case_id = len(self.signatures)
            self.signatures.append(signature)
            requires = []
            exprs = self.all_exprs
            for _t, expr, criterion in tests_for(signature, self):
                Ordering(self, expr).requires(requires)
                requires.append(expr)
                index_type = bitmap_index_type(self, expr)
                if index_type is not None:
                    if expr not in exprs:
                        exprs[expr] = 1
                        if always_testable(expr):
                            Ordering(self, expr).requires([])
                    index_type(self, expr).add_case(case_id, criterion)
        return super(IndexedEngine, self)._add_method(signature, rule)

    def _generate_code(self):
        smig = SMIGenerator(self.function)
        all_exprs = map(self.to_expression, self.all_exprs)
        for expr in all_exprs:
            smig.maybe_cache(expr)

        memo = dict([(expr, smig.action_id(expr)) for expr in all_exprs])
        return smig.generate(self.build_root(memo)).func_code

    def _full_reset(self):
        # Replace the entire engine with a new one
        Dispatching(self.function).create_engine(self.__class__)

    synchronized()
    def seed_bits(self, expr, cases):
        return BitmapIndex(self, expr).seed_bits(cases)

    synchronized()
    def reseed(self, expr, criterion):
        return BitmapIndex(self, expr).reseed(criterion)

    # Make build() a synchronized method
    build = synchronized(TreeBuilder.build.im_func)

    def build_root(self, memo):
        return self.build(
            to_bits([len(self.signatures)])-1, frozenset(self.all_exprs), memo
        )

    def best_expr(self, cases, exprs):
        return super(IndexedEngine, self).best_expr(
            list(from_bits(cases)), exprs
        )

    def build_node(self, expr, cases, remaining_exprs, memo):
        return memo[expr], predicate_node_for(
            self, expr, cases, remaining_exprs, memo
        )

    def selectivity(self, expr, cases):
        return BitmapIndex(self, expr).selectivity(cases)


    def to_expression(self, expr):
        return expr









    def build_leaf(self, cases, memo):
        action = self.rules.default_action
        signatures = self.signatures
        registry = self.registry
        for case_no in from_bits(cases):
            action = combine_actions(action, registry[signatures[case_no]])
        # No need to memoize here, since the combined action probably isn't
        # a meaningful key, and template-compiled methods are memoized at a
        # lower level anyway.
        return (0, compile_method(action, self))

when(bitmap_index_type,  (IndexedEngine, Truth))(lambda en,ex:TruthIndex)
when(predicate_node_for, (IndexedEngine, Truth))
def truth_node(builder, expr, cases, remaining_exprs, memo):
    dont_cares, seedmap = builder.seed_bits(expr, cases)
    return (    # True/false tuple for Truth
        builder.build(seedmap[True][0] | dont_cares, remaining_exprs, memo),
        builder.build(seedmap[False][0] | dont_cares, remaining_exprs, memo)
    )

when(bitmap_index_type,  (IndexedEngine, Identity))(lambda en,ex:PointerIndex)
when(predicate_node_for, (IndexedEngine, Identity))
def identity_node(builder, expr, cases, remaining_exprs, memo):
    dont_cares, seedmap = builder.seed_bits(expr, cases)
    return dict(
        [(seed, builder.build(inc|dont_cares, remaining_exprs, memo))
            for seed, (inc, exc) in seedmap.iteritems()]
    )

when(bitmap_index_type,  (IndexedEngine, Comparison))(lambda en,ex:RangeIndex)
when(predicate_node_for, (IndexedEngine, Comparison))
def range_node(builder, expr, cases, remaining_exprs, memo):
    dontcares, seedmap = builder.seed_bits(expr, cases)
    return split_ranges(
        dontcares, seedmap, lambda cases: builder.build(cases, remaining_exprs, memo)
    )

try: frozenset
except NameError: from core import frozenset


when(bitmap_index_type,  (IndexedEngine, type(None)))(value(None))

when(bitmap_index_type,  (IndexedEngine, (IsInstance, IsSubclass)))(
    value(TypeIndex)
)

when(predicate_node_for, (IndexedEngine, (IsInstance, IsSubclass)))
def class_node(builder, expr, cases, remaining_exprs, memo):
    dontcares, seedmap = builder.seed_bits(expr, cases)
    cache = {}
    def lookup_fn(cls):
        try:
            inc, exc = seedmap[cls]
        except KeyError:
            builder.reseed(expr, Class(cls))
            seedmap.update(builder.seed_bits(expr, cases)[1])
            inc, exc = seedmap[cls]
        cbits = dontcares | inc
        cbits ^= (exc & cbits)
        return cache.setdefault(cls, builder.build(cbits,remaining_exprs,memo))

    return cache, lookup_fn


abstract()
def type_to_test(typ, expr, engine):
    """Convert `typ` to a ``Test()`` of `expr` for `engine`"""

when(type_to_test, (type,))
when(type_to_test, (ClassType,))
def std_type_to_test(typ, expr, engine):
    return Test(IsInstance(expr), Class(typ))

when(type_to_test, (istype,))
def istype_to_test(typ, expr, engine):
    return Test(IsInstance(expr), typ)





when(tests_for, (istype(tuple), Engine))
def tests_for_tuple(ob, engine):
    for cls, arg in zip(ob, engine.argnames):
        yield type_to_test(cls, Local(arg), engine)

def always_testable(expr):
    """Is `expr` safe to evaluate in any order?"""
    return False

#when(always_testable, (IsSubclass,))   XXX might not be a class!

when(always_testable, ((Identity, Truth, Comparison, IsInstance),))
def testable_criterion(expr):
    return always_testable(expr.expr)

when(always_testable, ((Local, Const),))(value(True))


when(parse_rule, (IndexedEngine, basestring))
def _parse_string(engine, predicate, ctx, cls):
    b = CriteriaBuilder(engine.arguments, ctx.localdict, ctx.globaldict, __builtins__)
    expr = b.parse(predicate)
    bindings = b.bindings[0]
    if cls is not None and engine.argnames:
        cls = type_to_test(cls, engine.arguments[engine.argnames[0]], engine)
        expr = intersect(cls, expr)
    # XXX rewrap body for bindings here?  problem: CSE isn't ready
    # XXX so we'd have to make a temporary wrapper that gets replaced later. :(
    # XXX (the wrapper would just always recalc the values)
    # XXX Ugly bit at that point is that we're (slowly) generating code TWICE
    return Rule(
        maybe_bind(ctx.body, bindings), expr, ctx.actiontype, ctx.sequence
    )








def maybe_bind(func, bindings):
    """Apply expression bindings to arguments, if applicable"""

    if not bindings or not hasattr(func, 'func_code'):
        return func     # no bindings or not a function

    args, varargs, varkw, defaults = inspect.getargspec(func) 
    if not args or isinstance(args[0], basestring):
        return func # no args or first arg isn't a tuple

    for arg in args[0]:
        if not isinstance(arg, basestring):  # nested tuple arg, not a binding
            return func

    for arg in args[0]:
        if arg in bindings:
            for arg in args[0]:
                if arg not in bindings:
                    raise TypeError("Missing binding for %r" % arg)
            break
    else:
        return func     # none of the tuple args are in the binding

    argtuple = Tuple([bindings[arg] for arg in args[0]])

    c = Code.from_spec(func.func_name, args[1:], varargs, varkw)
    f = new.function(
        c.code(), func.func_globals, func.func_name, func.func_defaults
    )
    f.func_code = c.code()  # make f's signature match func w/out tuple
    c.return_(call_thru(f, Const(func), [argtuple]))    # call to match that
    f.func_code = c.code()  # now include the actual body
    f.__predicate_bindings__ = bindings, func   # mark for later optimization

    return f






# === As of this point, it should be possible to compile expressions!
#
meta_function(isinstance)(
    lambda __star__, __dstar__, *args, **kw:
        compileIsXCall(isinstance, IsInstance, args, kw, __star__, __dstar__)
)
meta_function(issubclass)(
    lambda __star__, __dstar__, *args, **kw:
        compileIsXCall(issubclass, IsSubclass, args, kw, __star__, __dstar__)
)
def compileIsXCall(func, test, args, kw, star, dstar):    
    if (
        kw or len(args)!=2 or not isinstance(args[1], Const)
        or isinstance(args[0], Const) or star is not None or dstar is not None
    ):
        return expressionSignature(
            Call(Const(func), args, tuple(kw.items()), star, dstar)
        )
    expr, seq = args
    return Test(test(expr), Disjunction(map(Class, _yield_tuples(seq.value))))

def _yield_tuples(ob):
    if type(ob) is tuple:
        for i1 in ob:
            for i2 in _yield_tuples(i1):
                yield i2
    else:
        yield ob

when(compileIs,
    # matches 'type(x) is y'
    "isinstance(expr,Call) and isinstance(expr.func, Const)"
    " and (expr.func.value is type) and len(expr.args)==1"
)
def compileTypeIsX(expr, criterion):
    return Test(IsInstance(expr.args[0]), istype(criterion))

when(expressionSignature, "isinstance(expr, Const) and isinstance(expr.value,priority)")
def test_for_priority(expr):
    return Test(None, expr.value)




import unittest, sys
from peak.rules.core import *

x2 = lambda a: a*2
x3 = lambda next_method, a: next_method(a)*3

class TypeEngineTests(unittest.TestCase):

    def testIntraSignatureCombinationAndRemoval(self):
        abstract()
        def f(a):
            """blah"""

        rx2 = Rule(x2,(int,), Method)
        rx3 = Rule(x3,(int,), Around)

        rules_for(f).add(rx2)
        self.assertEqual(f(1), 2)

        rules_for(f).add(rx3)
        self.assertEqual(f(1), 6)

        rules_for(f).remove(rx3)
        self.assertEqual(f(1), 2)

    def testAroundDecoratorAndRetroactiveCombining(self):
        def f(a):
            return a

        self.assertEqual(f(1), 1)
        self.assertEqual(f('x'), 'x')

        when(f, (int,))(x2)
        self.assertEqual(f(1), 2)
        self.assertEqual(f('x'), 'x')

        around(f, (int,))(lambda a:42)
        self.assertEqual(f(1), 42)
        self.assertEqual(f('x'), 'x')


class MiscTests(unittest.TestCase):

    def testPointers(self):
        from peak.rules.indexing import IsObject
        from sys import maxint
        anOb = object()
        ptr = IsObject(anOb)
        self.assertEqual(id(anOb)&maxint,ptr)
        self.assertEqual(hash(id(anOb)&maxint),hash(ptr))

        self.assertEqual(ptr.match, True)
        self.assertEqual(IsObject(anOb, False).match, False)
        self.assertNotEqual(IsObject(anOb, False), ptr)

        class X: pass
        anOb = X()
        ptr = IsObject(anOb)
        oid = id(anOb)&maxint
        self.assertEqual(oid,ptr)
        self.assertEqual(hash(oid),hash(ptr))
        del anOb
        self.assertNotEqual(ptr,"foo")
        self.assertEqual(ptr,ptr)
        self.assertEqual(hash(oid),hash(ptr))

    def testRuleSetReentrance(self):
        from peak.rules.core import Rule, RuleSet
        rs = RuleSet()
        log = []
        class MyListener:
            def actions_changed(self, added, removed):
                log.append(1)
                if self is ml1:
                    rs.unsubscribe(ml2)
        ml1, ml2 = MyListener(), MyListener()
        rs.subscribe(ml1)
        rs.subscribe(ml2)
        self.assertEqual(log, [])
        rs.add(Rule(lambda:None))
        self.assertEqual(log, [1, 1])

    def testAbstract(self):
        def f1(x,y=None):
            raise AssertionError("Should never get here")
        d = Dispatching(f1)
        log = []
        d.rules.default_action = lambda *args: log.append(args)
        f1 = abstract(f1)
        f1(27,42)
        self.assertEqual(log, [(27,42)])
        when(f1, ())(lambda *args: 99)
        self.assertRaises(AssertionError, abstract, f1)

    def testAbstractRegeneration(self):
        def f1(x,y=None):
            raise AssertionError("Should never get here")
        d = Dispatching(f1)
        log = []
        d.rules.default_action = lambda *args: log.append(args)
        d.request_regeneration()
        f1 = abstract(f1)
        self.assertNotEqual(d.backup, f1.func_code)
        self.assertEqual(f1.func_code, d._regen)
        f1.func_code = d.backup
        f1(27,42)
        self.assertEqual(log, [(27,42)])
        
    def testCreateEngine(self):
        def f1(x,y=None):
            raise AssertionError("Should never get here")
        d = Dispatching(f1)
        old_engine = d.engine
        self.assertEqual(d.rules.listeners, [old_engine])
        from peak.rules.core import TypeEngine
        class MyEngine(TypeEngine): pass
        d.create_engine(MyEngine)
        new_engine = d.engine
        self.assertNotEqual(new_engine, old_engine)
        self.failUnless(isinstance(new_engine, MyEngine))
        self.assertEqual(d.rules.listeners, [new_engine])


    def testIndexClassicMRO(self):
        class MyEngine: pass
        eng = MyEngine()
        from peak.rules.indexing import TypeIndex
        from peak.rules.criteria import Class
        from types import InstanceType
        ind = TypeIndex(eng, 'classes')
        ind.add_case(0, Class(MyEngine))
        ind.add_case(1, Class(object))
        ind.reseed(Class(InstanceType))
        self.assertEqual(
            dict(ind.expanded_sets()),
            {MyEngine: [[0,1],[]], InstanceType: [[1],[]], object: [[1],[]]}
        )

    def testEngineArgnames(self):
        argnames = lambda func: Dispatching(func).engine.argnames
        self.assertEqual(
            argnames(lambda a,b,c=None,*d,**e: None), list('abcde')
        )
        self.assertEqual(
            argnames(lambda a,b,c=None,*d: None), list('abcd')
        )
        self.assertEqual(
            argnames(lambda a,b,c=None,**e: None), list('abce')
        )
        self.assertEqual(
            argnames(lambda a,b,c=None: None), list('abc')
        )
        self.assertEqual(
            argnames(lambda a,(b,(c,d)), e: None), list('abcde')
        )

    def test_istype_implies(self):
        self.failUnless(implies(istype(object), object))
        self.failUnless(implies(istype(int), object))
        self.failIf(implies(istype(object, False), object))
        self.failIf(implies(istype(int, False), object))



    def testIndexedEngine(self):
        from peak.rules.predicates import IndexedEngine, Comparison
        from peak.rules.criteria import Range, Value, Test, Signature
        from peak.util.assembler import Local
        from peak.util.extremes import Min, Max
        abstract()
        def classify(age): pass
        Dispatching(classify).create_engine(IndexedEngine)
        def setup(r, f):
            when(classify, Signature([Test(Comparison(Local('age')), r)]))(f)
        setup(Range(hi=( 2,-1)), lambda age:"infant")
        setup(Range(hi=(13,-1)), lambda age:"preteen")
        setup(Range(hi=( 5,-1)), lambda age:"preschooler")
        setup(Range(hi=(20,-1)), lambda age:"teenager")
        setup(Range(lo=(20,-1)), lambda age:"adult")
        setup(Range(lo=(55,-1)), lambda age:"senior")
        setup(Value(16), lambda age:"sweet sixteen")

        self.assertEqual(classify(0),"infant")
        self.assertEqual(classify(25),"adult")
        self.assertEqual(classify(17),"teenager")
        self.assertEqual(classify(13),"teenager")
        self.assertEqual(classify(12.99),"preteen")
        self.assertEqual(classify(4),"preschooler")
        self.assertEqual(classify(55),"senior")
        self.assertEqual(classify(54.9),"adult")
        self.assertEqual(classify(14.5),"teenager")
        self.assertEqual(classify(16),"sweet sixteen")
        self.assertEqual(classify(16.5),"teenager")
        self.assertEqual(classify(99),"senior")
        self.assertEqual(classify(Min),"infant")
        self.assertEqual(classify(Max),"senior")

    def testSignatureOfTruthTests(self):
        from peak.rules.predicates import Truth        
        from peak.rules.criteria import Test, Signature
        # this line used to fail with a recursion error:
        Signature([Test(Truth(99), True), Test(Truth(88), False)])



    def testClassBodyRules(self):
        from peak.rules.core import Local, Rule
        from peak.rules.criteria import Signature, Test, Class, Value
        from peak.rules.predicates import IsInstance, Truth

        abstract()
        def f1(a): pass

        abstract()
        def f2(b): pass

        # This is to verify that the rules have sequence numbers by definition
        # order, not reverse definition order, inside a class.
        num = Rule(None).sequence
        
        class T:
            f1=sys._getframe(1).f_locals['f1']  # ugh
            when(f1)
            def f1_(a): pass

            f2=sys._getframe(1).f_locals['f2']
            when(f2, 'b')
            def f2_(b): pass

        self.assertEqual(
            list(rules_for(f1)), [Rule(T.f1_.im_func, (T,), Method, num+1)]
        )
        self.assertEqual(
            list(rules_for(f2)), [Rule(
                T.f2_.im_func, Signature([
                    Test(IsInstance(Local('b')), Class(T)),
                    Test(Truth(Local('b')), Value(True))
                ]), Method, num+2)
            ]
        )






    def testParseInequalities(self):
        from peak.rules.predicates import CriteriaBuilder, Comparison, Truth
        from peak.util.assembler import Compare, Local
        from peak.rules.criteria import Inequality, Test, Value
        from peak.rules.ast_builder import parse_expr
        builder = CriteriaBuilder(
            dict(x=Local('x'), y=Local('y')), locals(), globals(), __builtins__
        )
        pe = builder.parse

        x_cmp_y = lambda op, t=True: Test(
            Truth(Compare(Local('x'), ((op, Local('y')),))), Value(True, t)
        )      
        x,y = Comparison(Local('x')), Comparison(Local('y'))

        for op, mirror_op, not_op, stdop, not_stdop in [
            ('>', '<', '<=','>','<='),
            ('<', '>', '>=','<','>='),
            ('==','==','!=','==','!='),
            ('<>','<>','==','!=','=='),
        ]:
            fwd_sig = Test(x, Inequality(op, 1))
            self.assertEqual(pe('x %s 1' % op), fwd_sig)
            self.assertEqual(pe('1 %s x' % mirror_op), fwd_sig)

            rev_sig = Test(x, Inequality(mirror_op, 1))
            self.assertEqual(pe('x %s 1' % mirror_op), rev_sig)
            self.assertEqual(pe('1 %s x' % op), rev_sig)

            not_sig = Test(x, Inequality(not_op, 1))
            self.assertEqual(pe('not x %s 1' % op), not_sig)
            self.assertEqual(pe('not x %s 1' % not_op), fwd_sig)

            self.assertEqual(pe('x %s y' % op), x_cmp_y(stdop))
            self.assertEqual(pe('x %s y' % not_op), x_cmp_y(not_stdop))

            self.assertEqual(pe('not x %s y' % op),x_cmp_y(stdop,False))
            self.assertEqual(pe('not x %s y' % not_op),x_cmp_y(not_stdop,False))



    def testInheritance(self):
        class X: pass
        class Y(X): pass
        x = Y()
        f = lambda x: "f"
        when(f, "isinstance(x, X)")(lambda x: "g")
        when(f, "type(x) is X")(lambda x: "h")
        self.assertEqual(f(x), 'g')

    def testNotInherited(self):
        f = abstract(lambda x: "f")
        when(f, "not isinstance(x, int)")(lambda x: "g")
        when(f, "type(x) is object")(lambda x: "h")
        self.assertEqual(f(None), 'g')

    def testTypeImplicationAndIsSubclassOrdering(self):
        from inspect import isclass
        class A(object): pass
        class B(A): pass
        whats_this = abstract(lambda obj: obj)
        when(whats_this, "isclass(obj) and issubclass(obj, A)")(lambda o:"A")
        when(whats_this, "isclass(obj) and issubclass(obj, B)")(lambda o:"B")
        when(whats_this, (B,))(lambda o:"B()")
        when(whats_this, (A,))(lambda o:"A()")
        self.failUnlessEqual(whats_this(B), "B")
        self.failUnlessEqual(whats_this(A), "A")
        self.failUnlessEqual(whats_this(B()), "B()")
        self.failUnlessEqual(whats_this(A()), "A()")

    def testRuleSetClear(self):
        from peak.rules.core import Rule, RuleSet
        rs = RuleSet(); r = Rule(lambda:None,actiontype=Method)
        rs.add(r)
        self.assertEqual(list(rs), [r])
        rs.clear()
        self.assertEqual(list(rs), [])





    def testTypeEngineKeywords(self):
        def func(x, **k): pass
        def fstr(x,**k): return x,k
        abstract(func)
        when(func,(str,))(fstr)
        self.assertEqual(func('x',s='7'), ('x',{'s':'7'}))

    def testFlatPriorities(self):
        from peak.rules import value, AmbiguousMethods
        from peak.rules.predicates import priority
        
        f = lambda n, m: 0
        f1, f2, f3 = value(1), value(2), value(3)
        
        when(f, "n==5 and priority(1)")(f1)
        when(f, "m==5 and priority(1)")(f2)
        when(f, "n==5 and m==5 and priority(1)")(f3)
        self.assertEqual(f(5,5), 3)

    def testIsNot(self):
        def func(x): pass
        p,q,r = object(),object(),object()
        when(func,"x is not p")(value('~p'))
        when(func,"x is not p and x is not q and x is not r")(value('nada'))
        self.assertEqual(func(23),'nada')
        self.assertEqual(func(q),'~p')
        
    def testTypeVsIsTypePrecedence(self):
        def func(x): pass
        when(func, (int,        ))(value(1))
        when(func, (istype(int),))(value(2))
        self.assertEqual(func(42), 2)

    def testNamedGFExtension(self):
        p,q,r = object(),object(),object()
        when("%s:%s.named_func" % (__name__, self.__class__.__name__), "x is not p")(value('~p'))
        self.assertEqual(self.named_func(q),'~p')

    def named_func(x): pass
    named_func = staticmethod(named_func)

class RuleDispatchTests(unittest.TestCase):

    def testSimplePreds(self):
        from peak.rules import dispatch

        [dispatch.generic()]
        def classify(age):
            """Stereotype for age"""

        def defmethod(gf,s,func):
            gf.when(s)(func)

        defmethod(classify,'not not age<2', lambda age:"infant")
        defmethod(classify,'age<13', lambda age:"preteen")
        defmethod(classify,'age<5',  lambda age:"preschooler")
        defmethod(classify,'20>age', lambda age:"teenager")
        defmethod(classify,'not age<20',lambda age:"adult")
        defmethod(classify,'age>=55',lambda age:"senior")
        defmethod(classify,'age==16',lambda age:"sweet sixteen")
        self.assertEqual(classify(25),"adult")
        self.assertEqual(classify(17),"teenager")
        self.assertEqual(classify(13),"teenager")
        self.assertEqual(classify(12.99),"preteen")
        self.assertEqual(classify(0),"infant")
        self.assertEqual(classify(4),"preschooler")
        self.assertEqual(classify(55),"senior")
        self.assertEqual(classify(54.9),"adult")
        self.assertEqual(classify(14.5),"teenager")
        self.assertEqual(classify(16),"sweet sixteen")
        self.assertEqual(classify(16.5),"teenager")
        self.assertEqual(classify(99),"senior")
        self.assertEqual(classify(dispatch.strategy.Min),"infant")
        self.assertEqual(classify(dispatch.strategy.Max),"senior")








    def testKwArgHandling(self):
        from peak.rules import dispatch
        [dispatch.generic()]
        def f(**fiz): """Test of kw handling"""

        [f.when("'x' in fiz")]
        def f(**fiz): return "x"

        [f.when("'y' in fiz")]
        def f(**fiz): return "y"

        self.assertEqual(f(x=1),"x")
        self.assertEqual(f(y=1),"y")
        self.assertRaises(dispatch.AmbiguousMethod, f, x=1, y=1)

    def testVarArgHandling(self):
        from peak.rules import dispatch
        [dispatch.generic()]
        def f(*fiz): """Test of vararg handling"""

        [f.when("'x' in fiz")]
        def f(*fiz): return "x"

        [f.when("'y' in fiz")]
        def f(*fiz): return "y"

        self.assertEqual(f("foo","x"),"x")
        self.assertEqual(f("bar","q","y"),"y")
        self.assertEqual(f("bar","q","y"),"y")
        self.assertEqual(f("y","q",),"y")
        self.assertRaises(dispatch.AmbiguousMethod, f, "x","y")

    def test_NoApplicableMethods_is_raised(self):
        from peak.rules import dispatch
        [dispatch.generic()]
        def demo_func(number):
            pass
        demo_func.when("number < 10")(lambda x: 0)
        self.assertEqual(demo_func(3),0)
        self.assertRaises(dispatch.NoApplicableMethods, demo_func, 33)

    def testSingles(self):
        from peak.rules import dispatch        

        [dispatch.on('t')]
        def gm (t) : pass

        [gm.when(object)]
        def gm (t) : return 'default'

        [gm.when(int)]
        def gm2 (t) : return 'int'

        self.assertEqual(gm(42),"int")
        self.assertEqual(gm("x"),"default")
        self.assertEqual(gm(42.0),"default")


    def testSubclassDispatch(self):
        from peak.rules import dispatch        

        [dispatch.generic()]
        def gm (t) : pass

        [gm.when(dispatch.strategy.default)]
        def gm (t) : return 'default'

        [gm.when('issubclass(t,int)')]
        def gm2 (t) : return 'int'

        self.assertEqual(gm(int),"int")
        self.assertEqual(gm(object),"default")
        self.assertEqual(gm(float),"default")









    def testTrivialities(self):
        from peak.rules import dispatch        

        [dispatch.on('x')]
        def f1(x,*y,**z): "foo bar"

        [dispatch.on('x')]
        def f2(x,*y,**z): "baz spam"

        for f,doc in (f1,"foo bar"),(f2,"baz spam"):
            self.assertEqual(f.__doc__, doc)

            # Empty generic should raise NoApplicableMethods
            self.assertRaises(dispatch.NoApplicableMethods, f, 1, 2, 3)
            self.assertRaises(dispatch.NoApplicableMethods, f, "x", y="z")

            # Must have at least one argument to do dispatching
            self.assertRaises(TypeError, f)
            self.assertRaises(TypeError, f, foo="bar")






















class MockBuilder:
    def __init__(self, test, expr, cases, remaining, seeds, index=None):
        self.test = test
        self.args = expr, cases, remaining, {}
        self.seeds = seeds
        self.index = index

    def test_func(self, func):
        return func(self, *self.args)

    def build(self, cases, remaining_exprs, memo):
        self.test.failUnless(memo is self.args[-1])
        self.test.assertEqual(self.args[-2], remaining_exprs)
        return cases

    def seed_bits(self, expr, cases):
        self.test.assertEqual(self.args[1], cases)
        if self.index is not None:
            return self.index.seed_bits(cases)
        return self.seeds
        
    def reseed(self, expr, criterion):
        self.test.assertEqual(self.args[0], expr)
        self.index.reseed(criterion)

















class NodeBuildingTests(unittest.TestCase):

    def build(self, func, dontcare, seeds, index=None):
        seedbits = dontcare, seeds
        builder = MockBuilder(
            self, 'expr', 'cases', 'remaining', seedbits, index
        )
        return builder.test_func(func)

    def testTruthNode(self):
        from peak.rules.predicates import truth_node
        node = self.build(truth_node, 27,
            {True: (128,0), False: (64,0)})
        self.assertEqual(node, (27|128, 27|64))

    def testIdentityNode(self):
        from peak.rules.predicates import identity_node
        node = self.build(identity_node, 27,
            {9127: (128,0), 6499: (64,0), None: (0,0)})
        self.assertEqual(node, {None:27, 9127:27|128, 6499: 27|64})

    def testRangeNode(self):
        from peak.rules.indexing import RangeIndex, to_bits
        from peak.rules.predicates import range_node
        from peak.rules.criteria import Range, Value, Min, Max
        ind = RangeIndex(self, 'expr')
        ind.add_case(0, Value(19))
        ind.add_case(1, Value(23))
        ind.add_case(2, Value(23, False))
        ind.add_case(3, Range(lo=(57,1)))
        ind.add_case(4, Range(lo=(57,-1)))
        dontcare, seeds = ind.seed_bits(to_bits(range(6)))
        exact, ranges = self.build(range_node, dontcare, seeds)
        self.assertEqual(exact,
            {19:to_bits([0, 2, 5]), 23:to_bits([1,5]), 57:to_bits([2,4,5])})
        self.assertEqual(ranges,
            [((Min,57), to_bits([2,5])),  ((57,Max), to_bits([2,3,4,5]))])




    def testClassNode(self):
        from peak.rules.indexing import TypeIndex, to_bits
        from peak.rules.predicates import class_node
        from peak.rules.criteria import Class, Conjunction
        from types import InstanceType
        ind = TypeIndex(self, 'expr')
        class a: pass
        class b: pass
        class c(a,b): pass
        class x(a,b,object): pass

        ind.add_case(0, Class(InstanceType))
        ind.add_case(1, Conjunction([Class(a), Class(b), Class(c,False)]))
        ind.add_case(2, Class(object))
        ind.add_case(3, Conjunction([Class(a), Class(b)]))
        ind.add_case(4, Class(a))
        ind.selectivity(range(6))
        cases = to_bits(range(6))
        builder = MockBuilder(
            self, 'expr', cases, 'remaining', ind.seed_bits(cases), ind
        )
        cache, lookup = builder.test_func(class_node,)
        self.assertEqual(cache, {})
        data = [
            (object, to_bits([2,5])),
            (InstanceType, to_bits([0,2,5])),
            (a, to_bits([0,2,4,5])),
            (b, to_bits([0,2,5])),
            (c, to_bits([0,2,3,4,5])),
            (x, to_bits([1,2,3,4,5]))
        ]
        for k, v in data:
            self.assertEqual(lookup(k), v)
        self.assertEqual(cache, dict(data))
        
        





class DecompilationTests(unittest.TestCase):

    def setUp(self):
        from peak.rules.ast_builder import parse_expr
        from peak.rules.codegen import ExprBuilder        
        from peak.util.assembler import Local, Const
        from peak.rules.debug.decompile import decompile
        class Const(Const): pass    # non-folding
        class ExprBuilder(ExprBuilder):
            def Const(self,value): return Const(value)
        argmap = dict([(name,Local(name)) for name in 'abcdefgh'])
        builder = ExprBuilder(argmap, locals(), globals(), __builtins__)
        self.parse = builder.parse
        self.decompile = decompile

    def roundtrip(self, s1, s2=None):
        if s2 is None:
            s2 = s1
        self.check(self.parse(s1), s2)

    def check(self, expr, s):
        self.assertEqual(self.decompile(expr), s)

    def test_basics(self):
        from peak.util.assembler import Pass, Getattr, Local
        self.check(Pass, '')
        self.check(Getattr(Local('a'), 'b'), 'a.b')
        self.roundtrip('not -+~`a`')
        self.roundtrip('a+b-c*d/e%f**g//h')
        self.roundtrip('a and b or not c')
        self.roundtrip('~a|b^c&d<<e>>f')
        self.roundtrip('(a, b)')
        self.roundtrip('[a, b]')
        self.roundtrip('[a]')
        self.roundtrip('[]')
        self.roundtrip('()')
        self.roundtrip('1')
        self.roundtrip('a.b.c')
        # TODO: Compare


    def test_slices(self):
        self.roundtrip('a[:]')
        self.roundtrip('a[1:]')
        self.roundtrip('a[:2]')
        self.roundtrip('a[1:2]')
        self.roundtrip('a[1:2:3]')
        self.roundtrip('a[::]', 'a[:]')
        self.roundtrip('a[b:c:d]')
        self.roundtrip('a[b::d]')
        self.roundtrip('a[-b:c+d:e*f]')
        self.roundtrip('a[::d]')
        
    def test_paren_precedence(self):
        self.roundtrip('(1).__class__')
        self.roundtrip('1.2.__class__')
        self.roundtrip('`(a, b)`')
        self.roundtrip('`a,b`', '`(a, b)`')
        self.roundtrip('(a, [b, c])')
        self.roundtrip('a,b', '(a, b)')
        self.roundtrip('a+b*c')
        self.roundtrip('c*(a+b)')
        self.roundtrip('a*b*c')
        self.roundtrip('(a*b)*c', 'a*b*c')
        self.roundtrip('a*(b*c)')
        self.roundtrip('(a*b).c')
        self.roundtrip('(a, b, c)')
        self.roundtrip('a**b**c')
        self.roundtrip('a**(b**c)', 'a**b**c')
        self.roundtrip('a and b or c and not d')
        self.roundtrip('a and (b or c) and not d')

        if sys.version>='2.5':
            # if-else is right-associative
            self.roundtrip('(a, b, c) if d else e')
            self.roundtrip('(a if b else c) if d else e')
            self.roundtrip('a if (b if c else d) else e')
            self.roundtrip('a if b else (c if d else e)',
                           'a if b else c if d else e')



    def test_powers(self):
        # Exponentiation operator binds less tightly than unary numeric/bitwise
        # on the right:
        self.roundtrip('2**-1')
        self.roundtrip('2**+1')
        self.roundtrip('2**~1')
        self.roundtrip('2**1*2')
        self.roundtrip('2**(1*2)')
        
    def test_dicts(self):
        self.roundtrip('{}')
        self.roundtrip('{a: b}')
        self.roundtrip('{a: b, c: d}')

    def test_calls(self):
        self.roundtrip('1()')
        self.roundtrip('a()')
        self.roundtrip('a(b)')
        self.roundtrip('a(c=d)')
        self.roundtrip('a(*e)')
        self.roundtrip('a(**f)')
        self.roundtrip('a(b, c=d, *e, **f)')
        self.roundtrip('a.b(c)')
        self.roundtrip('a+b(c)')
        self.roundtrip('(a+b)(c)')
        if sys.version>='2.5':
            self.roundtrip('a if b(c) else d')
            self.roundtrip('a(b if c else d, *e if f else g)')





























def additional_tests():
    import doctest
    files = [
        'README.txt', 'DESIGN.txt', 'Indexing.txt', 'AST-Builder.txt',
        'Code-Generation.txt', 'Syntax-Matching.txt', 'Criteria.txt',
        'Predicates.txt', 
    ][sys.version<'2.4':]   # skip README.txt on 2.3 due to @ syntax
    return doctest.DocFileSuite(
        optionflags=doctest.ELLIPSIS|doctest.NORMALIZE_WHITESPACE, *files
    )


































from __future__ import division
import sys
from peak.util.addons import AddOn
from peak.rules.core import *
from peak.rules.criteria import *
from peak.rules.core import sorted, set, frozenset
from peak.util.extremes import Min, Max, Extreme
from peak.util.decorators import decorate
from types import InstanceType

__all__ = [
    'Ordering', 'BitmapIndex', 'TreeBuilder', 'split_ranges',
    'to_bits', 'from_bits', 'RangeIndex', 'TypeIndex', 'TruthIndex',
    'PointerIndex', 'bitmap_index_type'
]

def define_ordering(ob, seq):
    items = []
    for key in seq:
        Ordering(ob, key).requires(items)
        items.append(key)




















class Ordering(AddOn):
    """Track inter-expression ordering constraints"""

    def __init__(self, owner, expr):
        self.constraints = set()

    def requires(self, guards):
        c = frozenset(guards)
        cs = self.constraints
        if c in cs:
            return
        for oldc in list(cs):
            if c >= oldc:
                return  # already a less-restrictive condition
            elif oldc >= c:
                cs.remove(oldc)
        cs.add(c)

    def can_precede(self, exprs):
        if not self.constraints:
            return True
        for c in self.constraints:
            for e in c:
                if e in exprs:
                    break
            else:
                return True
        else:
            return False












def to_bits(ints):
    """Return a bitset encoding the numbers contained in sequence `ints`"""
    b = 0
    for i in ints:
        b |= 1<<i
    return b

if sys.version<"2.4":
    def to_bits(ints):
        """Return a bitset encoding the numbers contained in sequence `ints`"""
        b = 0
        for i in ints:
            if i>31:
                i = long(i)
            b |= 1<<i   # under 2.3, this breaks when i>31 unless it's a long
        return b

def from_bits(n):
    """Yield the (ascending) numbers contained in bitset n"""
    b = 0
    while n:
        while not n & 15:
            n >>= 4
            b += 4
        if n & 1:
            yield b
        n >>= 1
        b += 1













class TreeBuilder(object):
    """Template methods for the Chambers&Chen dispatch tree algorithm"""

    def build(self, cases, exprs, memo):
        key = (cases, exprs)
        if key in memo:
            return memo[key]
        if not exprs:
            node = self.build_leaf(cases, memo)
        else:
            best, rest = self.best_expr(cases, exprs)
            assert len(rest) < len(exprs)

            if best is None:
                # No best expression found, recurse
                node = self.build(cases, rest, memo)
            else:
                node = self.build_node(best, cases, rest, memo)
        memo[key] = node
        return node

    def build_node(self, expr, cases, remaining_exprs, memo):
        raise NotImplementedError

    def build_leaf(self, cases, memo):
        raise NotImplementedError

    def selectivity(self, expression, cases):
        """Return (seedcount,totalcases) selectivity statistics"""
        raise NotImplementedError

    def cost(self, expr, remaining_exprs):
        return 1








    def best_expr(self, cases, exprs):
        best_expr = None
        best_spread = None
        to_do = list(exprs)
        remaining = dict.fromkeys(exprs)
        active_cases = len(cases)
        skipped = []

        while to_do:
            expr = to_do.pop()
            if not Ordering(self, expr).can_precede(remaining):
                # Skip criteria that have unchecked prerequisites
                skipped.append(expr)
                continue

            branches, total_cases = self.selectivity(expr, cases)

            if total_cases == active_cases * branches:
                # None of the branches for this expression eliminate any
                # cases, so this expression isn't needed for dispatching
                del remaining[expr]

                # recheck skipped exprs that might be legal now
                to_do.extend(skipped)
                skipped = []
                continue

            spread = total_cases / branches
            if best_expr is None or spread < best_spread:
                best_expr, best_spread = expr, spread
                best_cost = self.cost(expr, remaining)
            elif spread==best_spread:
                cost = self.cost(expr, remaining)
                if cost < best_cost:
                    best_expr, best_cost = expr, cost

        if best_expr is not None:
            del remaining[best_expr]
        return best_expr, frozenset(remaining)


class BitmapIndex(AddOn):
    """Index that computes selectivity and handles basic caching w/bitmaps"""

    known_cases = 0
    match = True,

    decorate(staticmethod)
    def addon_key(*args):
        # Always use BitmapIndex as the add-on key
        return (BitmapIndex,)+args

    def __new__(cls, engine, expr):
        if cls is BitmapIndex:
            cls = bitmap_index_type(engine, expr)
        return super(BitmapIndex, cls).__new__(cls)

    def __init__(self, engine, expr):
        self.extra = {}
        self.null = self.all_seeds = {}    # seed -> inc_cri, exc_cri
        self.criteria_bits = {}             # cri  -> case bits
        self.case_seeds = []                # case -> seed set
        self.criteria_seeds = {}            # cri  -> seeds

    def add_case(self, case_id, criterion):
        if criterion in self.criteria_seeds:
            seeds = self.criteria_seeds[criterion]
        else:
            self.criteria_bits.setdefault(criterion, 0)
            seeds = self.criteria_seeds[criterion] = self.add_criterion(criterion)

        case_seeds = self.case_seeds
        if case_id==len(case_seeds):
            case_seeds.append(seeds)
        else:
            self._extend_cases(case_id)
            case_seeds[case_id] = seeds

        bit = to_bits([case_id])
        self.known_cases |= bit
        self.criteria_bits[criterion] |= bit

    def add_criterion(self, criterion):
        """Ensure `criterion` is indexed, return its "applicable seeds" set
        
        As a side effect, ``self.all_seeds`` must be updated to include any
        new seeds required by `criterion`.
        """
        raise NotImplementedError

    def _extend_cases(self, case_id):
        if case_id >= len(self.case_seeds):
            self.case_seeds.extend(
                [self.null]*(case_id+1-len(self.case_seeds))
            )

    def selectivity(self, cases):
        if cases and cases[-1] >= len(self.case_seeds):
            self._extend_cases(cases[-1])
        cases = map(self.case_seeds.__getitem__, cases)
        return (len(self.all_seeds), sum(map(len, cases)))

    def seed_bits(self, cases):
        bits = self.criteria_bits
        return cases ^ (self.known_cases & cases), dict([
            (seed,
                (sum([bits[i] for i in inc]) & cases,
                 sum([bits[e] for e in exc]) & cases))
                for seed, (inc, exc) in self.all_seeds.items()
        ])

    def expanded_sets(self):
        return [
            (seed, [list(from_bits(inc)), list(from_bits(exc))])
            for seed, (inc, exc) in self.seed_bits(self.known_cases)[1].items()
        ]

    def reseed(self, criterion):
        self.add_criterion(criterion)




    def include(self, seed, criterion, exclude=False):
        try:
            s = self.all_seeds[seed]
        except KeyError:
            s = self.all_seeds[seed] = set(), set()
        s[exclude].add(criterion)
        self.criteria_bits.setdefault(criterion, 0)

    def exclude(self, seed, criterion):
        self.include(seed, criterion, True)































def split_ranges(dont_cares, bitmap, node=lambda b:b):
    """Return (exact, ranges) where `exact` is a dict[value]->node and `ranges`
    is a sorted list of ``((lo,hi),node)`` tuples expressing non-inclusive
    ranges.  `dont_cares` and `bitmap` should be the return values from
    a bitmap index's ``seed_bits()`` method.  `node(bits)` should return
    the value to be used as a node in the output; the default is to just return
    a bitmap of cases.
    """
    ranges = []
    exact = {}
    current = dont_cares
    for (val,d), (inc, exc) in bitmap.iteritems():
        if d and not (val is Min and d==-1 and not inc):
            break     # something other than == was used; use full algorithm
        exact[val] = node(current | inc)
    else:
        return exact, [((Min, Max), node(dont_cares))]

    low = Min

    for (val,d), (inc, exc) in sorted(bitmap.iteritems()):
        if val != low:
            if ranges and ranges[-1][-1]==current:
                low = ranges.pop()[0][0]
            ranges.append(((low, val), node(current)))
            low = val
        new = current | inc
        new ^= (new & exc)
        if not isinstance(val, Extreme):
            if d==0 or d<0:
                exact[val] = node(new)
            elif val not in exact:
                exact[val] = node(current)
        if d:
            current = new
    if low != Max:
        if ranges and ranges[-1][-1]==current:
            low = ranges.pop()[0][0]
        ranges.append(((low, Max), node(current)))
    return exact, ranges

class PointerIndex(BitmapIndex):
    """Index for pointer equality"""

    def add_criterion(self, criterion):
        if isinstance(criterion, Intersection):
            match = False
            items = criterion
        else:
            match = criterion.match
            items = [criterion]
        for cri in items:
            seed = id(cri.ref)
            self.extra[seed] = 1
            if match:
                self.include(seed, criterion)
                self.exclude(None, criterion)
            else:
                self.include(None, criterion)
                self.exclude(seed, criterion)
        if match:
            return self.match
        return self.extra

    def seed_bits(self, cases):
        dontcares, seedmap = BitmapIndex.seed_bits(self, cases)
        defaults = seedmap[None][0] # default inclusions
        for seed, (inc, exc) in seedmap.items():
            if seed is not None:
                inc |= defaults ^ (defaults & exc)
                seedmap[seed] = inc, exc
        return dontcares, seedmap

class TruthIndex(BitmapIndex):
    """Index for simple boolean expression tests"""

    def add_criterion(self, criterion):
        assert isinstance(criterion, Value) and criterion.value is True
        self.include(criterion.match, criterion)
        self.include(not criterion.match, not criterion)
        return self.match

abstract()
def bitmap_index_type(engine, expr):
    """Get the BitmapIndex subclass to use for the given engine and `expr`"""
    raise NotImplementedError(engine, expr)

def class_seeds_for(criterion):
    """Yield class objects that 'criterion' thinks are relevant"""
    if isinstance(criterion, istype):
        yield criterion.type
    elif isinstance(criterion, Class):
        yield criterion.cls
    elif isinstance(criterion, Intersection):
        for c in criterion:
            for seed in class_seeds_for(c):
                yield seed

class TypeIndex(BitmapIndex):
    """Index for istype(), Class(), and Classes() criteria"""

    def add_class(self, cls):        
        t = istype(cls)
        for criterion, seeds in self.criteria_seeds.iteritems():
            if implies(t, criterion):
                self.include(cls, criterion)
                seeds.add(cls)
        self.include(cls, t)    # ensure it's in all_seeds

    def reseed(self, criterion):
        map(self.add_class, class_seeds_for(criterion))
        if object not in self.all_seeds:
            self.include(object, istype(object))

    def add_criterion(self, criterion):
        my_seeds = self.criteria_seeds.setdefault(criterion, set())
        self.reseed(criterion)
        for seed in self.all_seeds:
            if implies(istype(seed), criterion):
                self.include(seed, criterion)
                my_seeds.add(seed)
        return my_seeds

class RangeIndex(BitmapIndex):

    def __init__(self, engine, expr):
        BitmapIndex.__init__(self, engine, expr)
        self.null = None

    def add_criterion(self, criterion):

        if isinstance(criterion, Range):
            lo = i = criterion.lo
            hi = e = criterion.hi
            match = True

        elif isinstance(criterion, Value):
            lo = hi = val = (criterion.value, 0)
            if criterion.match:
                e = (Min, -1)
                i = val
            else:
                i = (Min, -1)
                e = val
            match = criterion.match

        else:
            raise TypeError(criterion)

        if (i not in self.all_seeds or e not in self.all_seeds
            or (lo,hi,match) not in self.extra
        ):
            # oops, there's a seed we haven't seen before:
            # make sure the offsets are rebuilt on the next selectivity() call
            self.extra.clear()   

        self.include(i, criterion)
        self.exclude(e, criterion)
        return lo, hi, match





    def selectivity(self, cases):
        case_seeds = self.case_seeds
        if cases and cases[-1] >= len(case_seeds):
            self._extend_cases(cases[-1])

        extras = self.extra
        if not extras:
            all_seeds = self.all_seeds
            offsets = dict([(k,n) for n,k in enumerate(sorted(all_seeds))])
            all = extras[None] = len(all_seeds)
            all_but_one = all - 1  
            for lo,hi,inc in self.criteria_seeds.values():
                if lo==hi:
                    extras[lo,hi,inc] = [all_but_one, 1][inc]
                else:
                    extras[lo,hi,inc] = offsets[hi] - offsets[lo]

        cases = map(case_seeds.__getitem__, cases)
        return (len(self.all_seeds), sum(map(extras.__getitem__, cases)))

























__all__ = [
    'Rule', 'RuleSet', 'Dispatching', 'Engine', 'rules_for', 'compile_method',
    'Method', 'Around', 'Before', 'After', 'MethodList', 'value',
    'DispatchError', 'AmbiguousMethods', 'NoApplicableMethods',
    'abstract', 'when', 'before', 'after', 'around', 'istype', 'parse_rule',
    'implies', 'dominant_signatures', 'combine_actions', 'overrides',
    'always_overrides', 'merge_by_default', 'intersect', 'disjuncts', 'negate'
]
from peak.util.decorators import decorate_assignment, decorate, struct, \
            synchronized, frameinfo, decorate_class, classy, apply_template
from peak.util.assembler import Code, Const, Call, Local, Getattr, TryExcept, \
            Suite, with_name
from peak.util.addons import AddOn
import inspect, new, itertools, operator, sys
try:
    set, frozenset = set, frozenset
except NameError:
    from sets import Set as set
    from sets import ImmutableSet
    class frozenset(ImmutableSet):
        """Kludge to fix the abomination that is ImmutableSet.__init__"""
        def __new__(cls, iterable=None):
            self = ImmutableSet.__new__(cls, iterable)
            ImmutableSet.__init__(self, iterable)
            return self
        def __init__(self, iterable=None):
            pass    # all immutable initialization should be done by __new__!
empty = frozenset()
try:
    sorted = sorted
except NameError:
    def sorted(seq,key=None):
        if key:
            d = [(key(v),v) for v in seq]
        else:
            d = list(seq)
        d.sort()
        if key:
            return [v[1] for v in d]
        return d

# Core logic -- many of these are generics to be specialized w/"when" later

def disjuncts(ob):
    """Return a *list* of the logical disjunctions of `ob`"""
    # False == no condition is sufficient == no disjuncts
    if ob is False: return []
    if type(ob) is tuple: return _tuple_disjuncts(ob)
    return [ob]

def implies(s1,s2):
    """Is s2 always true if s1 is true?"""
    return s1==s2

def overrides(a1, a2):
    """Does action a1 take precedence over action a2?"""
    return False

def combine_actions(a1,a2):
    """Return a new action for the combination of a1 and a2"""
    if a1 is None:
        return a2
    elif a2 is None:
        return a1
    elif overrides(a1,a2):
        if not overrides(a2,a1):
            return a1.override(a2)
    elif overrides(a2,a1):
        return a2.override(a1)
    return a1.merge(a2)

def rules_for(f):
    """Return the initialized ruleset for a generic function"""
    if not Dispatching.exists_for(f):
        d = Dispatching(f)
        d.rules.add(Rule(clone_function(f)))
    return Dispatching(f).rules





class Dispatching(AddOn):
    """Manage a generic function's rules, engine, locking, and code"""
    engine = None
    def __init__(self, func):
        self.function = func
        self._regen   = self._regen_code()  # callback to regenerate code
        self.rules    = RuleSet(self.get_lock())
        self.backup   = None  # allows func to call itself during regeneration
        self.create_engine(TypeEngine)

    synchronized()
    def get_lock(self):
        return self.__lock__

    def create_engine(self, engine_type):
        """Create a new engine of `engine_type`, unsubscribing old"""
        if self.engine is not None and self.engine in self.rules.listeners:
            self.rules.unsubscribe(self.engine)
        self.engine = engine_type(self)
        return self.engine

    synchronized()
    def request_regeneration(self):
        """Ensure code regeneration occurs on next call of the function"""
        if self.backup is None:
            self.backup = self.function.func_code
            self.function.func_code = self._regen

    def _regen_code(self):
        c = Code.from_function(self.function, copy_lineno=True)
        c.return_(
            call_thru(
                self.function,
                Call(Getattr(
                    Call(Const(Dispatching), (Const(self.function),), fold=False),
                    '_regenerate'
                ))
            )
        )
        return c.code()

    synchronized()
    def as_abstract(self):
        for action in self.rules:
            raise AssertionError("Can't make abstract: rules already exist")

        c = Code.from_function(self.function, copy_lineno=True)
        c.return_(call_thru(self.function, Const(self.rules.default_action)))

        if self.backup is None:
            self.function.func_code = c.code()
        else:
            self.backup = c.code()
        return self.function

    synchronized()
    def _regenerate(self):
        func = self.function
        assert self.backup is not None
        func.func_code = self.backup    # ensure re-entrant calls work

        try:
            # try to replace the code with new code
            func.func_code = self.engine._generate_code()
        except:
            # failure: we'll try to regen again, next time we're called
            func.func_code = self._regen
            raise
        else:
            # success!  get rid of the old backup code and return the function
            self.backup = None
            return func










class DispatchError(Exception):
    """A dispatch error has occurred"""

    def __call__(self,*args,**kw):
        raise self.__class__(*self.args+(args,kw))  # XXX

    def __repr__(self):
        # This method is needed so doctests for 2.3/2.4 match 2.5
        return self.__class__.__name__+repr(self.args)


class MethodType(type):
    """Metaclass for method types

    This allows precedence to be declared between method types, and ensures
    that ``compiled()`` methods aren't inherited when ``__call__`` is redefined
    in a Method subclass.
    """

    def __init__(cls, name, bases, cdict):
        if '__call__' in cdict and 'compiled' not in cdict:
            # Ensure 'compiled()' is not inherited for changed __call__
            cls.compiled = lambda self, engine: self
        return type.__init__(cls, name, bases, cdict)

    def __rshift__(self, other):
        if type(other) is tuple:
            for item in other:
                always_overrides(self, item)
        else:
            always_overrides(self, other)
        return other

    def __rrshift__(self, other):
        if type(other) is tuple:
            for item in other:
                always_overrides(item, self)
        else:
            always_overrides(other, self)
        return self

class Method(object):
    """A simple method w/optional chaining"""

    __metaclass__ = MethodType
    
    def __init__(self, body, signature=(), serial=0, tail=None):
        self.body = body
        self.signature = signature
        self.serial = serial
        self.tail = tail
        self.can_tail = False
        try:
            args = inspect.getargspec(body)[0]
        except TypeError:
            pass
        else:
            if args and args[0]=='next_method':
                if getattr(body, 'im_self', None) is None:  # already bound?
                    self.can_tail = True

    decorate(classmethod)
    def make(cls, body, signature=(), serial=0):
        return cls(body, signature, serial)

    def __repr__(self):
        data = (self.body, self.signature, self.serial, self.tail)
        return self.__class__.__name__+repr(data)

    def __call__(self, *args, **kw):
        """Slow way to call a method -- use compile_method instead!"""
        return compile_method(self)(*args, **kw)
        
    def override(self, other):
        if not self.can_tail:
            return self
        return self.tail_with(combine_actions(self.tail, other))

    def tail_with(self, tail):
        return self.__class__(self.body, self.signature, self.serial, tail)


    def merge(self, other):
        #if self.__class__ is other.__class__ and self.body is other.body:
        #    XXX need to merge signatures?
        #    return self.__class__(
        #        self.body, ???, ???, combine_actions(self.tail, other.tail)
        #    )
        return AmbiguousMethods([self,other])

    decorate(classmethod)
    def make_decorator(cls, name, doc=None):
        if doc is None:
            doc = "Extend a generic function with a method of type ``%s``" \
                  % cls.__name__
        if cls is Method:
            maker = None   # allow gf's to use something else instead of Method
        else:
            maker = cls.make
        def decorate(f, pred=(), depth=2, frame=None):
            def callback(frame, name, func, old_locals):
                assert f is not func    # XXX
                kind, module, locals_, globals_ = frameinfo(frame)
                context = ParseContext(func, maker, locals_, globals_, lineno)
                def register_for_class(cls, f=f):
                    _register_rule(f, pred, context, cls)
                    return cls
                if kind=='class':
                    # 'when()' in class body; defer adding the method
                    decorate_class(register_for_class, frame=frame)
                else:
                    register_for_class(None)
                if old_locals.get(name) is f:
                    return f    # prevent overwriting if name is the same
                return func
            rv = decorate_assignment(callback, depth, frame)
            if frame is None: frame = sys._getframe(depth-1)
            lineno = frame.f_lineno # this isn't valid w/out active trace!
            return rv
        decorate = with_name(decorate, name)
        decorate.__doc__ = doc
        return decorate

    def compiled(self, engine):
        body = compile_method(self.body, engine)
        if not self.can_tail:
            return body
        else:
            return new.instancemethod(body, compile_method(self.tail, engine))

when = Method.make_decorator(
    "when", "Extend a generic function with a new action"
)


class NoApplicableMethods(DispatchError):
    """No applicable action has been defined for the given arguments"""

    def merge(self, other):
        return AmbiguousMethods([self,other])


class AmbiguousMethods(DispatchError):
    """More than one choice of action is possible"""

    def __init__(self, methods, *args):
        DispatchError.__init__(self, methods, *args)
        mine = self.methods = []
        for m in methods:
            if isinstance(m, AmbiguousMethods):
                mine.extend(m.methods)
            else:
                mine.append(m)

    def merge(self, other):
        return AmbiguousMethods(self.methods+[other])

    def override(self, other):
        return self

    def __repr__(self):
        return "AmbiguousMethods(%s)" % self.methods


class RuleSet(object):
    """An observable, stably-ordered collection of rules"""
    default_action = NoApplicableMethods()
    default_actiontype = Method

    def __init__(self, lock=None):
        self.rules = []
        self.actiondefs = {}
        self.listeners = []
        if lock is not None:
            self.__lock__ = lock

    synchronized()
    def add(self, rule):
        actiondefs = frozenset(self._actions_for(rule))
        self.rules.append( rule )
        self.actiondefs[rule] = actiondefs
        self._notify(added=actiondefs)

    synchronized()
    def remove(self, rule):
        actiondefs = self.actiondefs.pop(rule)
        self.rules.remove(rule)
        self._notify(removed=actiondefs)

    #def changed(self, rule):
    #    sequence, actions = self.actions[rule]
    #    new_actions = frozenset(self._actions_for(rule, sequence))
    #    self.actions[rule] = sequence, new_actions
    #    self.notify(new_actions-actions, actions-new_actions)

    synchronized()
    def clear(self):
        actiondefs = frozenset(self)
        del self.rules[:]; self.actiondefs.clear()
        self._notify(removed=actiondefs)

    def _notify(self, added=empty, removed=empty):
        for listener in self.listeners[:]:  # must be re-entrant
            listener.actions_changed(added, removed)

    synchronized()
    def __iter__(self):
        ad = self.actiondefs
        return iter([a for rule in self.rules for a in ad[rule]])

    def _actions_for(self, (na, body, predicate, actiontype, seq)):
        actiontype = actiontype or self.default_actiontype
        for signature in disjuncts(predicate):
            yield Rule(body, signature, actiontype, seq)

    synchronized()
    def subscribe(self, listener):
        self.listeners.append(listener)
        if self.rules:
            listener.actions_changed(frozenset(self), empty)

    synchronized()
    def unsubscribe(self, listener):
        self.listeners.remove(listener)


def _register_rule(gf, pred, context, cls):
    """Register a rule for `gf` with possible import-deferring"""
    if not isinstance(gf, basestring):
        rules = rules_for(gf)
        rules.add(parse_rule(Dispatching(gf).engine, pred, context, cls))
        return
    if len(gf.split(':'))<>2 or len(gf.split())>1:
        raise TypeError(
            "Function specifier %r is not in 'module.name:attrib.name' format"
            % (gf,)
        )
    modname, attrname = gf.split(':')
    from peak.util.imports import whenImported
    def _delayed_register(module):
        for attr in attrname.split('.'):
            module = getattr(module, attr)
        _register_rule(module, pred, context, cls)
    whenImported(modname, _delayed_register)


class Engine(object):
    """Abstract base for dispatching engines"""

    reset_on_remove = True

    def __init__(self, disp):
        self.function = disp.function
        self.registry = {}
        self.closures = {}
        self.rules = disp.rules
        self.__lock__ = disp.get_lock()
        self.argnames = list(
            flatten(filter(None, inspect.getargspec(self.function)[:3]))
        )
        self.rules.subscribe(self)

    synchronized()
    def actions_changed(self, added, removed):
        if removed and self.reset_on_remove:
            return self._full_reset()
        for rule in removed:
            self._remove_method(rule.predicate, rule)
        for rule in added:
            self._add_method(rule.predicate, rule)
        if added or removed:
            self._changed()

    def _changed(self):
        """Some change to the rules has occurred"""
        Dispatching(self.function).request_regeneration()

    def _full_reset(self):
        """Regenerate any code, caches, indexes, etc."""
        self.registry.clear()
        self.actions_changed(self.rules, ())
        Dispatching(self.function).request_regeneration()





    compiled_cache = None

    def apply_template(self, template, *args):
        try:
            return self.compiled_cache[template, args]
        except (KeyError, TypeError, AttributeError):
            pass

        try:
            closure = self.closures[template]
        except KeyError:
            if template.func_closure:
                raise TypeError("Templates cannot use outer-scope variables")
            import linecache; from peak.util.decorators import cache_source
            tmp = apply_template(template, self.function, *args)
            body = ''.join(linecache.getlines(tmp.func_code.co_filename))
            filename = "<%s at 0x%08X wrapping %s at 0x%08X>" % (
                template.__name__, id(template),
                self.function.__name__, id(self)
            )
            d ={}
            exec compile(body, filename, "exec") in template.func_globals, d
            tmp, closure = d.popitem()
            closure.func_defaults = template.func_defaults
            cache_source(filename, body, closure)
            self.closures[template] = closure
        f = closure(self.function, *args)
        f.func_defaults = self.function.func_defaults

        try:
            hash(args)
        except TypeError:
            pass
        else:
            if self.compiled_cache is None:
                from weakref import WeakValueDictionary
                self.compiled_cache = WeakValueDictionary()
            self.compiled_cache[template, args] = f
        return f


    def _add_method(self, signature, rule):
        """Add a case for the given signature and rule"""
        registry = self.registry
        action = rule.actiontype(rule.body, signature, rule.sequence)
        if signature in registry:
            registry[signature] = combine_actions(registry[signature], action)
        else:
            registry[signature] = action
        return action

    def _remove_method(self, signature, rule):
        """Remove the case for the given signature and rule"""
        raise NotImplementedError

    def _generate_code(self):
        """Return a code object for the current state of the function"""
        raise NotImplementedError
























class TypeEngine(Engine):
    """Simple type-based dispatching"""

    cache = None

    def __init__(self, disp):
        self.static_cache = {}
        super(TypeEngine, self).__init__(disp)

    def _changed(self):
        if self.cache != self.static_cache:
            Dispatching(self.function).request_regeneration()

    def _bootstrap(self):
        # Bootstrap a self-referential generic function by ensuring an exact
        # list of signatures is always in the function's dispatch cache.
        #
        # Only peak.rules.core generic functions used in the implementation of
        # other generic functions need this; currently that's just implies()
        # and overrides(), which control method order and combining.
        #
        cache = self.static_cache
        for sig, act in self.registry.items():
            for key in type_keys(sig):
                cache[key] = compile_method(act, self)
        self._changed()

    def _add_method(self, signature, rule):
        action = super(TypeEngine, self)._add_method(signature, rule)
        cache = self.static_cache
        for key in cache.keys():
            if key==signature or implies(key, signature):
                del cache[key]
        return action







    def _generate_code(self):
        self.cache = cache = self.static_cache.copy()
        def callback(*args, **kw):
            types = tuple([getattr(arg,'__class__',type(arg)) for arg in args])
            key = tuple(map(istype, types))
            self.__lock__.acquire()
            try:
                action = self.rules.default_action
                for sig in self.registry:
                    if implies(key, sig):
                        action = combine_actions(action, self.registry[sig])
                f = cache[types] = compile_method(action, self)
            finally:
                self.__lock__.release()
            return f(*args, **kw)

        c = Code.from_function(self.function, copy_lineno=True)
        types = [class_or_type_of(Local(name))
                    for name in flatten(inspect.getargspec(self.function)[0])]
        target = Call(Const(cache.get), (tuple(types), Const(callback)))
        c.return_(call_thru(self.function, target))
        return c.code()


# Handle alternates in tuple signatures
#
# when(disjuncts, (istype(tuple),)) - this has to be hardcoded for bootstrap
def _tuple_disjuncts(ob):
    for posn, item in enumerate(ob):
        if type(item) is tuple:
            head = ob[:posn]
            return [
                head+(mid,)+tail
                    for tail in _tuple_disjuncts(ob[posn+1:]) for mid in item
            ]
    return [ob]     # no alternates





# Code generation stuff

def flatten(v):
    if isinstance(v,basestring): yield v; return
    for i in v:
        for ii in flatten(i): yield ii

def gen_arg(v):
    if isinstance(v,basestring): return Local(v)
    if isinstance(v,list): return tuple(map(gen_arg,v))

def call_thru(sigfunc, target, prefix=()):
    args, star, dstar, defaults = inspect.getargspec(sigfunc)
    return Call(target, list(prefix)+map(gen_arg,args), (), gen_arg(star), gen_arg(dstar), fold=False)

def class_or_type_of(expr):
    return Suite([expr, TryExcept(
        Suite([Getattr(Code.DUP_TOP, '__class__'), Code.ROT_TWO, Code.POP_TOP]),
        [(Const(AttributeError), Call(Const(type), (Code.ROT_TWO,)))]
    )])

def clone_function(f):
    return new.function(
      f.func_code, f.func_globals, f.func_name, f.func_defaults, f.func_closure
    )

_default_engine = None
def compile_method(action, engine=None):
    """Convert `action` into an optimized callable for `engine`"""
    if engine is None:
        global _default_engine
        if _default_engine is None:
            _default_engine = Dispatching(abstract(lambda *a, **k: None)).engine
        # allow any rules on non-None engines to apply
        return compile_method(action, _default_engine)
    if isinstance(action, (Method, value)):
        return action.compiled(engine)
    elif action is None:
        return engine.rules.default_action
    return action

# Rules management

def abstract(func=None):
    """Declare a function to be abstract"""
    if func is None:
        return decorate_assignment(
            lambda f,n,func,old: Dispatching(func).as_abstract()
        )
    else:
        return Dispatching(func).as_abstract()

next_sequence = itertools.count().next

struct()
def Rule(body, predicate=(), actiontype=None, sequence=None):
    if sequence is None:
        sequence = next_sequence()
    return body, predicate, actiontype, sequence

struct()
def ParseContext(
    body, actiontype=None, localdict=(), globaldict=(), lineno=None, sequence=None
):
    """Hold information needed to parse a predicate"""
    if sequence is None:
        sequence = next_sequence()
    return body, actiontype, dict(localdict), dict(globaldict), lineno, sequence

def parse_rule(engine, predicate, context, cls):
    """Hook for pre-processing predicates, e.g. parsing string expressions"""
    if cls is not None and type(predicate) is tuple:
        predicate = (cls,) + predicate        
    return Rule(context.body, predicate, context.actiontype, context.sequence)








# Class/type rules and implications

[struct()]
def istype(type, match=True):
    return type, match

def type_key(arg):
    if isinstance(arg, (type, ClassType)):
        return arg
    elif type(arg) is istype and arg.match:
        return arg.type

def type_keys(sig):
    if type(sig) is tuple:
        key = tuple(map(type_key, sig))
        if None not in key:
            yield key

when(implies, (istype(tuple), istype(tuple)))
def tuple_implies(s1,s2):
    if len(s2)>len(s1):
        return False    # shorter tuple can't imply longer tuple
    for t1,t2 in zip(s1,s2):
        if not implies(t1,t2):
            return False
    else:
        return True

from types import ClassType, InstanceType
when(implies, (type,      (ClassType, type) ))(issubclass)
when(implies, (ClassType,  ClassType        ))(issubclass)
when(implies, (istype,     istype           ))(lambda s1,s2:
    s1==s2 or (s1.type is not s2.type and s1.match and not s2.match))
when(implies, (istype,    (ClassType, type) ))(lambda s1,s2:
    s1.match and implies(s1.type,s2))

# A classic class only implies a new-style one if it's ``object``
# or ``InstanceType``; this is an exception to the general rule that
# isinstance(X,Y) implies issubclass(X.__class__,Y)
when(implies, (ClassType, type))(lambda s1,s2: s2 is object or s2 is InstanceType)

# Rule precedence

[struct(
    __call__ = lambda self,*a,**kw: self.value,
    compiled = lambda self, engine:
                      engine.apply_template(value_template, self.value),
    __repr__ = lambda self: 'value(%r)' % self[1:]
)]
def value(value):
    """Method body returning a constant value"""
    return value,

def value_template(__func,__value):
    return "return __value"

YES, NO = value(True), value(False)

def always_overrides(a, b):
    """`a` instances always override `b`s; `b` instances never override `a`s"""
    a,b = istype(a), istype(b)
    pairs = {}
    to_add = [(a,b)]
    for rule in rules_for(overrides):
        sig = rule.predicate
        if type(sig) is not tuple or len(sig)!=2 or rule.body is not YES:
            continue
        pairs[sig]=1
        if sig[0]==b: to_add.append((a, sig[1]))
        if sig[1]==a: to_add.append((sig[0], b))
    for (p1,p2) in to_add:
        if (p2,p1) in pairs:
            raise TypeError("%r already overrides %r" % (b.type, a.type))
    for (p1,p2) in to_add:
        if (p1,p2) not in pairs:
            when(overrides, (p1, p2))(YES)
            when(overrides, (p2, p1))(NO)

def merge_by_default(t):
    """instances of `t` never imply other instances of `t`"""
    when(overrides, (t, t))(NO)

class MethodList(Method):
    """A list of related methods"""

    can_tail = False
    _sorted_items = None

    def __init__(self, items=(), tail=None):
        self.items = list(items)
        self.tail = tail

    decorate(classmethod)
    def make(cls, body, signature=(), serial=0):
        return cls( [(serial, signature, body)] )

    def __repr__(self):
        data = self.items, self.tail
        return self.__class__.__name__+repr(data)

    def tail_with(self, tail):
        return self.__class__(self.items, tail)

    def merge(self, other):
        if other.__class__ is not self.__class__:
            raise TypeError("Incompatible action types for merge", self, other)
        return self.__class__(
            self.items+other.items, combine_actions(self.tail, other.tail)
        )

    def compiled(self, engine):
        wrappers = tuple(engine.rules.methodlist_wrappers)
        bodies = [compile_method(body,engine) for sig, body in self.sorted()]
        return engine.apply_template(list_template, tuple(bodies), wrappers)









    def sorted(self):
        if self._sorted_items is not None:
            return self._sorted_items

        self.items.sort()
        rest = [(s,b) for (serial, s, b) in self.items]

        self._sorted_items = items = []
        seen = set()
        while rest:
            best = dominant_signatures(rest)
            map(rest.remove, best)
            for s,b in best:
                if b not in seen:
                    seen.add(b)
                    items.append((s,b))
        return items

def list_template(__func, __bodies, __wrappers):
    return """
    def __iterate():
        for __body in __bodies: yield __body($args)
    __result = __iterate()
    for __wrapper in __wrappers:
        __result = __wrapper(__result)
    return __result"""

merge_by_default(MethodList)


class Around(Method):
    """'Around' Method (takes precedence over regular methods)"""

around = Around.make_decorator('around')







class Before(MethodList):
    """Method(s) to be called before the primary method(s)"""
    can_tail = True

    def compiled(self, engine):
        tail = compile_method(self.tail, engine)
        bodies = [compile_method(body,engine) for sig, body in self.sorted()]
        return engine.apply_template(before_template, tail, tuple(bodies))

def before_template(__func, __tail, __bodies):
    return """
    for __body in __bodies: __body($args)
    return __tail($args)"""
    
before = Before.make_decorator('before')

class After(MethodList):
    """Method(s) to be called after the primary method(s)"""
    can_tail = True

    def sorted(self):
        # Reverse the sorting for after methods
        if self._sorted_items is not None:
            return self._sorted_items
        items = super(After,self).sorted()
        items.reverse()
        return items

    def compiled(self, engine):
        tail = compile_method(self.tail, engine)
        bodies = [compile_method(body, engine) for sig, body in self.sorted()]
        return engine.apply_template(after_template, tail, tuple(bodies))

def after_template(__func, __tail, __bodies):
    return """
    __retval = __tail($args)
    for __body in __bodies: __body($args)
    return __retval"""

after  = After.make_decorator('after')

# Define the overall method order
Around >> Before >> After >> (Method, MethodList)

# These are necessary to ensure that any added Method subclasses will
# automatically override NoApplicableMethods (and any subclasses thereof):
#
when(overrides, (Method, NoApplicableMethods))(YES)
when(overrides, (NoApplicableMethods, Method))(NO)

when(overrides, (Method,Method))
def method_overrides(a1, a2):
    if a1.__class__ is a2.__class__:
        return implies(a1.signature, a2.signature)
    raise TypeError("Incompatible action types", a1, a2)

when(overrides, (AmbiguousMethods, Method))
def ambiguous_overrides(a1, a2):
    for m in a1.methods:
        if overrides(m, a2):
            # if any ambiguous method overrides a2, we can toss it
            return True
    return False

when(overrides, (Method, AmbiguousMethods))
def override_ambiguous(a1, a2):
    for m in a2.methods:
        if not overrides(a1, m):
            return False
    return True     # can only override if it overrides all the ambiguity

# needed to disambiguate the above two methods if combining a pair of AM's:
merge_by_default(AmbiguousMethods)

# And now we can bootstrap the core!  These two functions are used in the
# TypeEngine implementation, so we force them to statically cache all their
# current methods.  That way, they'll still work even if they're called during
# one of their own cache misses or code regenerations:
Dispatching(implies).engine._bootstrap()
Dispatching(overrides).engine._bootstrap()


when(parse_rule, (TypeEngine, istype(tuple, False)))
def parse_upgrade(engine, predicate, context, cls):
    """Upgrade to predicate dispatch engine when called w/unrecognized args"""
    if isinstance(predicate, (type, ClassType, istype)):
        # convert single item to tuple - no need to upgrade engine
        return parse_rule(engine, (predicate,), context, cls)
    from peak.rules.predicates import IndexedEngine
    return parse_rule(
        Dispatching(engine.function).create_engine(IndexedEngine),
        predicate, context, cls
    )

when(rules_for, type(After.sorted))(lambda f: rules_for(f.im_func))


# Logical functions needed for extensions to the core, but that should be
# shared by all extensions.

abstract()
def negate(c):
    """Return the logical negation of criterion `c`"""

when(negate, (bool,)  )(operator.not_)
when(negate, (istype,))(lambda c: istype(c.type, not c.match))

abstract()
def intersect(c1, c2):
    """Return the logical intersection of two conditions"""

around(intersect, (object, object))
def intersect_if_implies(next_method, c1, c2):
    if implies(c1,c2):      return c1
    elif implies(c2, c1):   return c2
    return next_method(c1, c2)

# These are needed for boolean intersects to work correctly
when(implies, (bool, bool))(lambda c1, c2: c2 or not c1)
when(implies, (bool, object))(lambda c1, c2: not c1)
when(implies, (object, bool))(lambda c1, c2: c2)


def dominant_signatures(cases):
    """Return the most-specific ``(signature,body)`` pairs from `cases`

    `cases` is a sequence of ``(signature,body)`` pairs.  This routine checks
    the ``implies()`` relationships between pairs of signatures, and then
    returns a list of ``(signature,method)`` pairs such that no signature
    remaining in the original list implies a signature in the new list.
    The relative order of cases in the new list is preserved.
    """

    if len(cases)==1:
        # Shortcut for common case
        return list(cases)

    best, rest = list(cases[:1]), list(cases[1:])

    for new_sig, new_meth in rest:

        for old_sig, old_meth in best[:]:   # copy so we can modify inplace

            new_implies_old = implies(new_sig, old_sig)
            old_implies_new = implies(old_sig, new_sig)

            if new_implies_old:

                if not old_implies_new:
                    # better, remove the old one
                    best.remove((old_sig, old_meth))

            elif old_implies_new:
                # worse, skip adding the new one
                break
        else:
            # new_sig has passed the gauntlet, as it has not been implied
            # by any of the current "best" items
            best.append((new_sig,new_meth))

    return best






"""RuleDispatch Emulation API"""
from peak.util.decorators import decorate_assignment
from peak.rules import core
from warnings import warn
import new, sys

__all__ = [
    'on', 'as', 'generic', 'as_'
]

def generic(combiner=None):
    """Please switch to using @peak.rules.abstract()"""
    if combiner is not None:
        raise AssertionError(
            "Custom combiners are not supported by the compatibility API;"
            " please use a custom method type instead"
        )
    def callback(frm,name,value,old_locals):
        value.when   = core.when.__get__(value)
        value.around = core.around.__get__(value)
        value.before = core.before.__get__(value)
        value.after  = core.after.__get__(value)
        def clear(): core.rules_for(value).clear()
        value.clear = clear
        return core.abstract(value)

    return decorate_assignment(callback)


def as_(*decorators):
    """Please switch to using @peak.util.decorators.decorate"""
    warn("Please use @peak.util.decorators.decorate instead of dispatch.as()",
        DeprecationWarning, 2
    )
    def callback(frame,k,v,old_locals):
        for d in decorators[::-1]: v = d(v)
        return v
    return decorate_assignment(callback)

globals()['as'] = as_  

def make_module():
    def callback(frm, name, value, old_locals):
        m = new.module('dispatch.'+name)
        v = value()
        m.__dict__.update(v)
        m.__all__ = list(v)
        sys.modules.setdefault(m.__name__, m)
        sys.modules.setdefault('peak.rules.'+m.__name__, m)        
        return m
    return decorate_assignment(callback)

make_module()
def interfaces():
    from peak.rules.core import DispatchError, NoApplicableMethods
    from peak.rules.core import AmbiguousMethods as AmbiguousMethod

    __all__.extend(locals())    # add these to the main namespace, too
    globals().update(locals())
    return locals()

make_module()
def strategy():
    default = ()
    from peak.util.extremes import Min, Max
    return locals()
















def on(argument_name):
    """Please switch to using @peak.rules.abstract()"""

    def callback(frm,name,value,old_locals):
        import inspect
        funcname = value.__name__
        args, varargs, kwargs, defaults = inspect.getargspec(value)
        if argument_name not in args:
            raise NameError("%r does not have an argument named %r"
                % (value, argument_name))
        argpos = args.index(argument_name)

        func = value

        def when(cond):
            """Add following function to this GF, using 'cond' as a guard"""
            def callback(frm,name,value,old_locals):
                core.when(func, (object,)*argpos + (cond,))(value)
                if old_locals.get(name) is func:
                    return func
                return value
            return decorate_assignment(callback)

        def clone():
            raise AssertionError("PEAK-Rules can't clone() generic functions")

        func.when = when
        func.clone = clone
        return core.abstract(func)

    return decorate_assignment(callback)


# Install
sys.modules.setdefault('dispatch', sys.modules[__name__])









from peak.util.assembler import *
from peak.util.symbols import Symbol
from peak.rules.core import gen_arg, clone_function
from ast_builder import build, parse_expr
import sys
try:
    set
except NameError:
    from sets import Set as set

__all__ = [
    'GetSlice', 'BuildSlice', 'Dict', 'ExprBuilder', 'IfElse', 'CSECode',
    'UnaryOp', 'BinaryOp', 'ListOp',
]

nodetype(fmt='%s[%s:%s]')
def GetSlice(expr, start=Pass, stop=Pass, code=None):
    if code is None:
        if expr is not Pass:
            return fold_args(GetSlice, expr, start, stop)
        return expr, start, stop
    code(expr)
    if start is not Pass:
        code(start)
        if stop is not Pass:
            return code(stop, Code.SLICE_3)
        return code.SLICE_1()
    elif stop is not Pass:
        code(stop)
        return code.SLICE_2()
    return code.SLICE_0()










nodetype(fmt="%s:%s:%s")
def BuildSlice(start=Pass, stop=Pass, stride=Pass, code=None):
    if code is None:
        return fold_args(BuildSlice, start, stop, stride)
    if start is Pass: start = None
    if stop  is Pass: stop  = None
    code(start, stop, stride)
    if stride is not Pass:
        return code.BUILD_SLICE(3)
    return code.BUILD_SLICE(2)

nodetype()
def Dict(items, code=None):
    if code is None:
        return fold_args(Dict, tuple(map(tuple,items)))
    code.BUILD_MAP(0)
    for k,v in items:
        code.DUP_TOP()
        code(k, v)
        code.ROT_THREE()
        code.STORE_SUBSCR()

nodetype(fmt='%s if %s else %s')
def IfElse(tval, cond, fval, code=None):
    if code is None:
        return fold_args(IfElse, tval, cond, fval)
    else_clause, end_if = Label(), Label()
    code(cond)
    if tval != cond:
        code(else_clause.JUMP_IF_FALSE_OR_POP, tval)
        if code.stack_size is not None:
            code(end_if.JUMP_FORWARD)
    elif fval != cond:
        code(end_if.JUMP_IF_TRUE)

    if fval !=cond:       
        return code(else_clause, Code.POP_TOP, fval, end_if)
    else:
        return code(else_clause, end_if)


def unaryOp(name, (fmt, opcode)):
    if '%' not in fmt: fmt += '%s'
    nodetype(UnaryOp, fmt=fmt)
    def tmp(expr, code=None):
        if code is None:
            return fold_args(tmp, expr)
        return code(expr, opcode)
    tmp.__name__ = name
    return tmp

def binaryOp(name, (fmt, opcode)):
    if '%' not in fmt: fmt = '%s' + fmt + '%s'
    nodetype(BinaryOp, fmt=fmt)
    def tmp(left, right, code=None):
        if code is None:
            return fold_args(tmp, left, right)
        return code(left, right, opcode)
    tmp.__name__ = name
    return tmp

def listOp(name, (fmt, opcode)):
    nodetype(ListOp, fmt=fmt)
    def tmp(items, code=None):
        if code is None:
            return fold_args(tmp, tuple(items))
        code(*items)
        return opcode(code, len(items))
    tmp.__name__ = name
    return tmp

def mkOps(optype, **ops):
    return dict([(name,optype(name, op)) for (name, op) in ops.items()])

def globalOps(optype, **ops):
    __all__.extend(ops)
    localOps(globals(), optype, **ops)

def localOps(ns, optype, **ops):
    ns.update(mkOps(optype, **ops))


class UnaryOp(object):
    __slots__ = ()

class BinaryOp(object):
    __slots__ = ()

class ListOp(object):
    __slots__ = ()

globalOps(
    unaryOp,
    Not = ('not ', Code.UNARY_NOT),
    Plus = ('+', Code.UNARY_POSITIVE),
    Minus = ('-', Code.UNARY_NEGATIVE),
    Repr = ('`%s`', Code.UNARY_CONVERT),
    Invert = ('~', Code.UNARY_INVERT),
)

globalOps(
    binaryOp,
    Add = ('+', Code.BINARY_ADD),
    Sub = ('-', Code.BINARY_SUBTRACT),
    Mul = ('*', Code.BINARY_MULTIPLY),
    Div = ('/', Code.BINARY_DIVIDE),
    Mod = ('%s%%%s', Code.BINARY_MODULO),
    FloorDiv = ('//', Code.BINARY_FLOOR_DIVIDE),
    Power = ('**', Code.BINARY_POWER),
    LeftShift = ('<<', Code.BINARY_LSHIFT),
    RightShift = ('>>', Code.BINARY_RSHIFT),
    Getitem = ('%s[%s]', Code.BINARY_SUBSCR),
    Bitor = ('|', Code.BINARY_OR),
    Bitxor = ('^', Code.BINARY_XOR),
    Bitand = ('&', Code.BINARY_AND),
)

globalOps(
    listOp, Tuple = ('(%s)', Code.BUILD_TUPLE), List = ('[%s]', Code.BUILD_LIST)
)



class SMIGenerator:
    """State Machine Interpreter Generator"""
    
    ARG = Local('$Arg')
    SET_ARG = lambda self, code: code.STORE_FAST(self.ARG.name)
    WHY_CONTINUE = {'2.3':5}.get(sys.version[:3], 32)

    def __init__(self, func):
        import inspect
        self.code = code = CSECode.from_function(func) #, copy_lineno=True)
        self.actions = {}
        self.func = func
        loop_top, exit, bad_action, fake = Label(), Label(), Label(), Label()
        args, star, dstar, defaults = inspect.getargspec(func)
        actions, self.actions_const = self.make_const({})
        start_node, self.startnode_const = self.make_const(object())

        code.cache(None)    # force CSE preamble here
        code(start_node, loop_top)
        code.UNPACK_SEQUENCE(2)     # action, argument
        code(
            exit.JUMP_IF_FALSE,
            Compare(Code.DUP_TOP, (('in', actions),)),
            bad_action.JUMP_IF_FALSE_OR_POP,
            Code.ROT_TWO,   # argument, action
            self.SET_ARG,   # action
            fake.SETUP_LOOP, self.WHY_CONTINUE, Code.END_FINALLY,
            Code.POP_BLOCK, fake, Return(Pass),  # <- all dead code, never runs
        exit,
            Code.POP_TOP,       # drop action, leaving argument
            Return(
                Call(Pass, map(gen_arg, args),(),gen_arg(star), gen_arg(dstar))
            ),
        bad_action,
            Code.POP_TOP,
            Return(Call(Const(self.bad_action),(Code.ROT_THREE, Code.ROT_TWO)))
        )
        self.NEXT_STATE = loop_top.JUMP_ABSOLUTE
        self.maybe_cache = code.maybe_cache


    def generate(self, start_node):
        func = clone_function(self.func)
        self.code.co_consts[self.startnode_const] = start_node
        self.code.co_consts[self.actions_const] = dict.fromkeys(
            self.actions.values()
        )
        func.func_code = self.code.code()
        return func

    def make_const(self, value):
        self.code.co_consts.append(value)
        return Const(value), len(self.code.co_consts)-1

    def action_id(self, expression):
        try:
            return self.actions[expression]
        except KeyError:
            action = self.actions[expression] = self.code.here()
            self.code.stack_size = 0
            self.code(expression)
            if self.code.stack_size is not None:
                self.code(self.NEXT_STATE)
            return action

    def bad_action(self, action, argument):
        raise AssertionError("Invalid action: %s, %s" % (action, argument))















CACHE = Local('$CSECache')
SET_CACHE = lambda code: code.STORE_FAST(CACHE.name)

class CSETracker(Code):
    """Helper object that tracks common sub-expressions"""

    def __init__(self):
        super(CSETracker, self).__init__()
        self.cse_depends = {}

    def track(self, expr):
        self.track_stack = [None, 0]
        self.to_cache = []
        try:
            self(expr)
            return self.to_cache
        finally:
            del self.track_stack, self.to_cache

    def __call__(self, *args):
        scall = super(CSETracker, self).__call__
        ts = self.track_stack
        for ob in args:           
            ts[-1] += 1
            ts.append(ob)
            ts.append(0)
            try:
                before = self.stack_size
                scall(ob)
            finally:
                count = ts.pop()
                ts.pop()
            if count and callable(ob) and self.stack_size==before+1:
                # Only consider non-leaf callables for caching
                top = tuple(ts[-2:])
                if self.cse_depends.setdefault(ob, top) != top:
                    if ob not in self.to_cache:
                        self.to_cache.append(ob)



class CSECode(Code):
    """Code object with common sub-expression caching support"""

    def __init__(self):
        super(CSECode, self).__init__()
        self.expr_cache = {}
        self.tracker = CSETracker()
        
    def cache(self, expr):
        if not self.expr_cache:
            self.LOAD_CONST(None)
            self.STORE_FAST(CACHE.name)
        self.expr_cache.setdefault(
            expr, "%s #%d" % (expr, len(self.expr_cache)+1)
        )

    def maybe_cache(self, expr):
        map(self.cache, self.tracker.track(expr))

    def __call__(self, *args):
        scall = super(CSECode, self).__call__
        for ob in args:
            if callable(ob) and ob in self.expr_cache:
                key = self.expr_cache[ob]
                def calculate(code):
                    scall(ob, Code.DUP_TOP, CACHE, Const(key), Code.STORE_SUBSCR)
                cache = IfElse(
                    CACHE, CACHE, lambda c: scall({}, Code.DUP_TOP, SET_CACHE)
                )
                scall(
                    IfElse(
                        Getitem(CACHE, Const(key)),
                        Compare(Const(key), [('in', cache)]),
                        calculate
                    )
                )
            else:
                scall(ob)



class ExprBuilder:
    """Expression builder returning bytecode-able AST nodes"""

    def __init__(self,arguments,*namespaces):
        self.bindings = [
            dict([(k,self.Const(v)) for k,v in ns.iteritems()]) for ns in namespaces
        ]
        self.push(arguments); self.push()

    def push(self, ns={}): self.bindings.insert(0, {}); self.bind(ns)
    def bind(self, ns): self.bindings[0].update(ns)
    def pop(self): return self.bindings.pop(0)        
    def parse(self, expr): return parse_expr(expr, self)
    def Const(self,value): return Const(value)

    def Name(self,name):
        for ns in self.bindings:
            if name in ns: return ns[name]
        raise NameError(name)

    def Subscript(self, left, right):
        expr = build(self,left)
        key =  build(self,right)
        if isinstance(key, GetSlice):
            return GetSlice(expr, key.start, key.stop)
        return Getitem(expr, key)

    def Slice2(self, start, stop):
        start = start and build(self, start) or Pass
        stop  = stop  and build(self, stop ) or Pass
        return GetSlice(Pass, start, stop)

    def Slice3(self, start, stop, stride):
        start  = start  and build(self, start ) or Pass
        stop   = stop   and build(self, stop  ) or Pass
        stride = stride and build(self, stride) or Pass
        return BuildSlice(start, stop, stride)

    def Getattr(self, expr, attr):
        return Getattr(build(self,expr), attr)

    simplify_comparisons = False

    def Compare(self, expr, ops):
        return Compare(
            build(self, expr),
            [(op=='<>' and '!=' or op, build(self,arg)) for op, arg in ops]
        )

    def _unaryOp(name, nt):
        def method(self, expr):
            return nt(build(self,expr))
        return method

    localOps(locals(), _unaryOp,
        UnaryPlus  = Plus,
        UnaryMinus = Minus,
        Invert     = Invert,
        Backquote  = Repr,
        Not        = Not,
    )

    del _unaryOp

    def _mkBinOp(name, nt):
        def method(self, left, right):
            return nt(build(self,left), build(self,right))
        return method

    localOps(locals(), _mkBinOp,
        Add        = Add,
        Sub        = Sub,
        Mul        = Mul,
        Div        = Div,
        Mod        = Mod,
        FloorDiv   = FloorDiv,
        Power      = Power,
        LeftShift  = LeftShift,
        RightShift = RightShift,
    )
    del _mkBinOp

    def _multiOp(name, nt):
        def method(self, items):
            result = build(self,items[0])
            for item in items[1:]:
                result = nt(result, build(self,item))
            return result
        return method

    localOps(locals(), _multiOp,
        Bitor  = Bitor,
        Bitxor = Bitxor,
        Bitand = Bitand,
    )
    del _multiOp

    def _listOp(name, op):
        def method(self,items):
            return op(map(build.__get__(self), items))
        return method

    localOps(locals(), _listOp,
        And   = And,
        Or    = Or,
        Tuple = Tuple,
        List  = List,
    )

    def Dict(self, items):
        b = build.__get__(self)
        return Dict([(b(k),b(v)) for k,v in items])

    def CallFunc(self, func, args, kw, star_node, dstar_node):
        b = build.__get__(self)
        return Call(
            b(func), map(b,args), [(b(k),b(v)) for k,v in kw],
            star_node and b(star_node), dstar_node and b(dstar_node)
        )

    def IfElse(self, tval, cond, fval):
        return IfElse(build(self,tval), build(self,cond), build(self,fval))




from peak.util.assembler import *
from codegen import *
from criteria import *
from predicates import *
from core import *
from token import NAME
from ast_builder import build

__all__ = ['match', 'Bind', 'match_predicate', 'match_sequence']

nodetype()
def Bind(name, code=None):
    if code is None:
        return name,
    raise TypeError("Can't compile Bind expression")

def match_predicate(pattern, expr, binds):
    """Return predicate matching pattern to expr, updating binds w/bindings"""
    return Test(Comparison(expr), Inequality('==', pattern))

when(match_predicate, (type(None),))
def match_none(pattern, expr, binds):
    return Test(Identity(expr), IsObject(pattern))

when(match_predicate, (Bind,))
def match_bind(pattern, expr, binds):
    if pattern.name != '_':
        vals = binds.setdefault(pattern.name, [])
        if expr not in vals:
            vals.append(expr)
            for old in vals[-2:-1]:
                return Test(Truth(Compare(expr, (('==', old),))), True)
    return True

class SyntaxBuilder(ExprBuilder):
    def Backquote(self, expr):
        while len(expr)==2: expr, expr = expr
        if expr[0]==NAME:
            return Bind(expr[1])
        raise SyntaxError("backquotes may only be used around an indentifier")

def match(expr, pattern):
    """Match `expr` against inline pattern `pattern`

    This function can only be called inside a rule string; the second argument
    is treated as a syntactic match pattern, which can include backquoted
    locals to be used as bind patterns."""
    raise NotImplementedError(
        "Use match_predicate() to match compiled patterns outside a rule"
    )

def build_pattern(builder, node):
    old = builder.__class__
    builder.__class__ = SyntaxBuilder
    try:
        return build(builder, node)
    finally:
        builder.__class__ = old

meta_function(match, pattern=build_pattern)
def compile_match(__builder__, expr, pattern):
    binds = {}
    pred = match_predicate(pattern, expr, binds)
    __builder__.bind(dict([(k, list(v)[0]) for k, v in binds.items()]))
    return pred

















when(match_predicate, (istype(list),))
when(match_predicate, (istype(tuple),))
def match_sequence(pattern, expr, binds):
   pred = Test(Comparison(Call(Const(len), (expr,))), Value(len(pattern)))
   for pos, item in enumerate(pattern):
       pred = intersect(
           pred, match_predicate(item, Getitem(expr, Const(pos)), binds)
       )
   return pred

when(match_predicate, (Node,))
def match_node(pattern, expr, binds):
   pred = Test(IsInstance(expr), istype(type(pattern)))
   for pos, item in enumerate(pattern):
       if pos:
           pred = intersect(
               pred, match_predicate(item, Getitem(expr, Const(pos)), binds)
           )
   return pred

























"""The PEAK Rules Framework"""

from peak.rules.core import abstract, when, before, after, around, istype, \
    DispatchError, AmbiguousMethods, NoApplicableMethods, value

def combine_using(*wrappers):
    """Designate a generic function that wraps the iteration of its methods

    Standard "when" methods will be combined by iteration in precedence order,
    and the resulting iterator will be passed to the supplied wrapper(s), last
    first.  (e.g. ``combine_using(sorted, itertools.chain)`` will chain the
    sequences supplied by each method into one giant list, and then sort it).

    As a special case, if you include ``abstract`` in the wrapper list, it
    will be removed, and the decorated function will be marked as abstract.

    This decorator can only be used once per function, and can't be used if
    the generic function already has methods (even the default method) or if
    a custom method type has already been set (e.g. if you already called
    ``combine_using()`` on it before).
    """
    is_abstract = abstract in wrappers
    if is_abstract:
        wrappers = tuple([w for w in wrappers if w is not abstract])
        
    def callback(frame, name, func, old_locals):
        if core.Dispatching.exists_for(func) and list(core.rules_for(func)):
            raise RuntimeError("Methods already defined for", func)
        if is_abstract:
            func = abstract(func)
        r = core.Dispatching(func).rules
        if r.default_actiontype is not core.Method:
            raise RuntimeError("Method type already defined for", func)
        r.default_actiontype = core.MethodList.make
        r.methodlist_wrappers = wrappers[::-1]
        if not is_abstract:
            r.add(core.Rule(core.clone_function(func)))
        return func
    return core.decorate_assignment(callback)


def expand_as(predicate_string):
    """In rules, use the supplied condition in place of the decorated function

    Usage::
        @expand_as('filter is None or value==filter')
        def check(filter, value):
            "Check whether value matches filter"

    When the above function is used in a rule, the supplied condition will
    be "inlined", so that PEAK-Rules can expand the function call without
    losing its ability to index the rule or determine its precedence relative
    to other rules.

    When the condition is inlined, it's done in a namespace-safe manner.
    Names in the supplied condition will refer to either the arguments of the
    decorated function, or to locals/globals in the frame where ``expand_as``
    was called.  Names defined in the condition body (e.g. via ``let``) will
    not be visible to the caller.

    To prevent needless code duplication, you do not need to provide a body for
    your function, unless it is to behave differently than the supplied
    condition when it's called outside a rule.  If the decorated function has
    no body of its own (i.e. it's a ``pass`` or just a docstring), the supplied
    condition will be compiled to provide one automatically.  (That's why the
    above example usage has no body for the ``check()`` function.)
    """    
    def callback(frame, name, func, old_locals):
        from peak.rules.predicates import _expand_as
        kind, module, locals_, globals_ = core.frameinfo(frame)
        return _expand_as(
            func, predicate_string, locals_, globals_, __builtins__
        )
    return core.decorate_assignment(callback)








def let(**kw):
    """Define temporary variables for use in rules and methods

    Usage::

        @when(somefunc, "let(x=foo(y), z=y*2) and x>z")
        def some_method((x,z), next_method, y):
            # do something here

    The keywords used in the let() expression become available for use in
    any part of the rule that is joined to the ``let()`` by an ``and``
    expression, but will not be available in expressions joined by ``or`` or
    ``not`` branches.  Any ``let()`` calls at the top level of the expression
    will also be available for use in the method body, if you place them in
    a tuple argument in the *very first* argument position -- even before
    ``next_method`` and ``self``.

    Note that variables defined by ``let()`` are **lazy** - their values are
    not computed until/unless they are actually needed by the relevant part
    of the rule, so it does not slow things down at runtime to list all your
    variables up front.  Likewise, only the variables actually listed in your
    first-argument tuple are calculated, and only when the method is actually
    invoked.

    (Currently, this feature is mainly to support easy-to-understand rules,
    and DRY method bodies, as variables used in the rule's criteria may be
    calculated a second time when the method is invoked.)

    Note that while variable calculation is lazy, there *is* an evaluation
    order *between* variables in a let; you can't use a let-variable before
    it's been defined; you'll instead get whatever argument, local, or global
    variable would be shadowed by the as-yet-undefined variable.
    """
    raise NotImplementedError("`let` can only be used in rules, not code!")

__all__ = [_k for _k in list(globals()) if not _k.startswith('_') and _k!='core']
# TEMPORARY BACKWARDS COMPATIBILITY - PLEASE IMPORT THIS DIRECTLY FROM CORE
# (or better still, use the '>>' operator that method types now have)
#
from peak.rules.core import *  # always_overrides, Method, etc.




from peak.rules.core import *
from peak.rules.core import sorted, frozenset, set
from peak.util.decorators import struct
from weakref import ref
from sys import maxint
from peak.util.extremes import Min, Max
__all__ = [
    'Range', 'Value', 'IsObject', 'Class', 'tests_for',
    'Conjunction', 'Disjunction', 'Test', 'Signature',
    'Inequality', 'DisjunctionSet', 'OrElse', 'Intersection',
]

class Intersection(object):
    """Abstract base for conjunctions and signatures"""
    __slots__ = ()

class Disjunction(object):
    """Abstract base for DisjunctionSet and OrElse
    
    Note that a Disjunction can never have less than 2 members, as creating a
    Disjunction with only 1 item returns that item, and creating one with no
    items returns ``False`` (as no acceptable conditions means "never true").
    """
    __slots__ = ()

    def __new__(cls, input):
        if cls is Disjunction:
            return DisjunctionSet(input)
        return super(Disjunction, cls).__new__(cls)

when(negate, (Disjunction,))(lambda c: reduce(intersect, map(negate, c)))
        
struct()
def Range(lo=(Min,-1), hi=(Max,1)):
    if hi<=lo: return False
    return lo, hi

struct()
def Value(value, match=True):
    return value, match

when(implies, (Value, Range))(
    # ==Value implies Range if Value is within range
    lambda c1, c2: c1.match and c2.lo <= (c1.value, 0) <= (c2.hi)
)
when(implies, (Range, Value))(
    # Range implies !=Value if Value is out of range
    lambda c1, c2: not c2.match and not (c1.lo <= (c2.value, 0) <= (c1.hi)) or
                   c2.match and c1==Range((c2.value,-1), (c2.value,1))
)
when(implies, (Range, Range))(
    # Range1 implies Range2 if both points are within Range2's bounds
    lambda c1, c2: c1.lo >= c2.lo and c1.hi <= c2.hi
)
when(implies, (Value, Value))(
    lambda c1, c2: c1==c2 or (c1.match and not c2.match and c1.value!=c2.value)
)
when(intersect, (Range, Range))(
    lambda c1, c2: Range(max(c1.lo, c2.lo), min(c1.hi, c2.hi))
)

def to_range(v):
    lo, hi = (v.value, -1), (v.value, 1)
    if v.match:
        return Range(lo, hi)
    return Disjunction([Range(hi=lo), Range(lo=hi)])

when(intersect, (Value, Value))
def intersect_values(c1, c2):
    if c1==c2: return c1
    return intersect(to_range(c1), to_range(c2))

# if a value is a point inside a range, the implication rules would handle it;
# therefore, they are either !=values, or outside the range (and thus empty)
when(intersect, (Range, Value))(
    lambda c1, c2: not c2.match and intersect(c1, to_range(c2)) or False
)
when(intersect, (Value, Range))(
    lambda c1, c2: not c1.match and intersect(to_range(c1), c2) or False
)


when(negate, (Value,))(lambda c: Value(c.value, not c.match))
when(negate, (Range,))(lambda c: Disjunction([Range(hi=c.lo), Range(lo=c.hi)]))


struct()
def Class(cls, match=True):
    return cls, match

when(negate, (Class,))(lambda c: Class(c.cls, not c.match))

when(implies, (Class, Class))
def class_implies(c1, c2):
    if c1==c2:
        # not isinstance(x) implies not isinstance(x) always
        #     isinstance(x) implies isintance(x)      always
        return True
    elif c1.match and c2.match:
        #     isinstance(x) implies     isinstance(y) if issubclass(x,y)
        return implies(c1.cls, c2.cls)
    elif c1.match or c2.match:
        # not isinstance(x) implies     isinstance(x) never
        #     isinstance(x) implies not isinstance(y) never
        return False
    else:
        # not isinstance(x) implies not isinstance(y) if issubclass(y, x)
        return implies(c2.cls, c1.cls)

when(intersect, (Class, Class))(lambda c1,c2: Conjunction([c1, c2]))

when(implies, (istype, Class))(lambda c1,c2:
    c1.match and (c2.match == implies(c1.type, c2.cls)) # use ob/inst rules
)
when(implies, (Class, istype))(lambda c1,c2:
    c1.match and not c2.match and c1.cls is not c2.type
    and issubclass(c1.cls, c2.type)
)





struct()
def Test(expr, criterion):
    d = disjuncts(criterion)
    if len(d)!=1:
        return Disjunction([Test(expr, c) for c in d])
    return expr, criterion

when(negate, (Test,))(lambda c: Test(c.expr, negate(c.criterion)))

when(implies, (Test, Test))(
    lambda c1, c2: c1.expr==c2.expr and implies(c1.criterion, c2.criterion)
)

when(intersect, (Test, Test))
def intersect_tests(c1, c2):
    if c1.expr==c2.expr:
        return Test(c1.expr, intersect(c1.criterion, c2.criterion))
    else:
        return Signature([c1, c2])

inequalities = {
    '>':  lambda v: Range(lo=(v, 1)),
    '>=': lambda v: Range(lo=(v,-1)),
    '<':  lambda v: Range(hi=(v,-1)),
    '<=': lambda v: Range(hi=(v, 1)),
    '!=': lambda v: Value(v, False),
    '==': lambda v: Value(v, True),
}

inequalities['=<'] = inequalities['<=']
inequalities['=>'] = inequalities['>=']
inequalities['<>'] = inequalities['!=']

def Inequality(op, value):
    return inequalities[op](value)






class Signature(Intersection, tuple):
    """Represent and-ed Tests, in sequence"""

    __slots__ = ()

    def __new__(cls, input):
        output = []
        index = {}
        input = iter(input)
        for new in input:
            if new is True:
                continue
            elif new is False:
                return False
            assert isinstance(new, Test), \
                "Signatures can only contain ``criteria.Test`` instances"
            expr = new.expr
            if expr in index:
                posn = index[expr]
                old = output[posn]
                if old==new:
                    continue    # 'new' is irrelevant, skip it
                new = output[posn] = intersect(old, new)
                if new is False: return False
            else:
                index[expr] = len(output)
                output.append(new)

        if not output:
            return True
        elif len(output)==1:
            return output[0]
        return tuple.__new__(cls, output)

    def __repr__(self):
        return "Signature("+repr(list(self))+")"

when(negate, (Signature,))(lambda c: OrElse(map(negate, c)))



class IsObject(int):
    """Criterion for 'is' comparisons"""

    __slots__ = 'ref', 'match'

    def __new__(cls, ob, match=True):
        self = IsObject.__base__.__new__(cls, id(ob)&maxint)
        self.match = match
        self.ref = ob
        return self

    def __eq__(self,other):
        return self is other or (
            isinstance(other, IsObject) and self.match==other.match
            and self.ref is other.ref
        ) or (isinstance(other,(int,long)) and int(self)==other and self.match)

    def __ne__(self,other):
        return not (self==other)

    def __repr__(self):
        return "IsObject(%r, %r)" % (self.ref, self.match)

when(negate, (IsObject,))(lambda c: IsObject(c.ref, not c.match))

when(implies, (IsObject, IsObject))
def implies_objects(c1, c2):
    # c1 implies c2 if it's identical, or if c1=="is x" and c2=="is not y"
    return c1==c2 or (c1.match and not c2.match and c1.ref is not c2.ref)

when(intersect, (IsObject, IsObject))
def intersect_objects(c1, c2):
    #  'is x and is y'            'is not x and is x'
    if (c1.match and c2.match) or (c1.ref is c2.ref):
        return False
    else:
        # "is not x and is not y"
        return Conjunction([c1,c2])



class DisjunctionSet(Disjunction, frozenset):
    """Set of minimal or-ed conditions (i.e. no redundant/implying items)

    Note that a Disjunction can never have less than 2 members, as creating a
    Disjunction with only 1 item returns that item, and creating one with no
    items returns ``False`` (as no acceptable conditions means "never true").
    """
    def __new__(cls, input):
        output = []
        for item in input:
            for new in disjuncts(item):
                for old in output[:]:
                    if implies(new, old):
                        break
                    elif implies(old, new):
                        output.remove(old)
                else:
                    output.append(new)
        if not output:
            return False
        elif len(output)==1:
            return output[0]
        return frozenset.__new__(cls, output)

when(implies, (Disjunction, (object, Disjunction)))
def union_implies(c1, c2):  # Or(...) implies x if all its disjuncts imply x
    for c in c1:
        if not implies(c, c2):
            return False
    else:
        return True

when(implies, (object, Disjunction))
def ob_implies_union(c1, c2):   # x implies Or(...) if it implies any disjunct
    for c in c2:
        if implies(c1, c):
            return True
    else:
        return False


# We use @around for these conditions in order to avoid redundant implication
# testing during intersection, as well as to avoid ambiguity with the
# (object, bool) and  (bool, object) rules for intersect().
#
around(intersect, (Disjunction, object))(
    lambda c1, c2: type(c1)([intersect(x,c2) for x in c1])
)
around(intersect, (object, Disjunction))(
    lambda c1, c2: type(c2)([intersect(c1,x) for x in c2])
)
around(intersect, (Disjunction, Disjunction))(
    lambda c1, c2: type(c1)([intersect(x,y) for x in c1 for y in c2])
)
around(intersect, (Disjunction, Intersection))(
    lambda c1, c2: type(c1)([intersect(x,c2) for x in c1])
)
around(intersect, (Intersection, Disjunction))(
    lambda c1, c2: type(c2)([intersect(c1,x) for x in c2])
)

# XXX These rules prevent ambiguity with implies(object, bool) and
# (bool, object), at the expense of redundancy.  This can be cleaned up later
# when we allow cloning of actions for an existing rule.  (That is, when we can
# say "treat (bool, Disjunction) like (bool, object)" without duplicating the
# action.)
#
when(implies, (bool, Disjunction))(lambda c1, c2: not c1)
when(implies, (Disjunction, bool))(lambda c1, c2: c2)

# The disjuncts of a Disjunction are a list of its contents:
when(disjuncts, (DisjunctionSet,))(list)

abstract()
def tests_for(ob, engine=None):
    """Yield the tests composing ob, if any"""

when(tests_for, (Test,     ))(lambda ob, e: iter([ob]))
when(tests_for, (Signature,))(lambda ob, e: iter(ob))
when(tests_for, (bool,     ))(lambda ob, e: iter([]))


class Conjunction(Intersection, frozenset):
    """Set of minimal and-ed conditions (i.e. no redundant/implied items)

    Note that a Conjunction can never have less than 2 members, as
    creating a Conjunction with only 1 item returns that item, and
    creating one with no items returns ``True`` (since no required conditions
    means "always true").
    """
    def __new__(cls, input):
        output = []
        for new in input:
            for old in output[:]:
                if implies(old, new):
                    break
                elif implies(new, old):
                    output.remove(old)
                elif mutually_exclusive(new, old):
                    return False
            else:
                output.append(new)
        if not output:
            return True
        elif len(output)==1:
            return output[0]
        return frozenset.__new__(cls, output)

around(implies, (Intersection, object))
def set_implies(c1, c2):
    for c in c1:
        if implies(c, c2): return True
    else:
        return False

around(implies, ((Intersection, object), Intersection))
def ob_implies_set(c1, c2):
    for c in c2:
        if not implies(c1, c): return False
    else:
        return True


when(negate, (Conjunction,))(lambda c: Disjunction(map(negate, c)))

when(intersect, (istype,Class))
def intersect_type_class(c1, c2):
    if not c1.match: return Conjunction([c1,c2])
    return False

when(intersect, (Class, istype))
def intersect_type_class(c1, c2):
    if not c2.match: return Conjunction([c1,c2])
    return False

when(intersect, (istype,istype))
def intersect_type_type(c1, c2):
    # This is only called if there's no implication
    if c1.match or c2.match:
        return False    # so unless both are False matches, it's an empty set
    return Conjunction([c1, c2])

def mutually_exclusive(c1, c2):
    """Is the intersection of c1 and c2 known to be always false?"""
    return False

when(mutually_exclusive, (istype, istype))(
    lambda c1, c2: c1.match != c2.match and c1.type==c2.type
)















# Intersecting an intersection with something else should return a set of the
# same type as the leftmost intersection.  These methods are declared @around
# to avoid redundant implication testing during intersection, as well as to
# avoid ambiguity with the (object, bool) and  (bool, object) rules for
# intersect().
#
around(intersect, (Intersection, Intersection))(
    lambda c1, c2: type(c1)(list(c1)+list(c2))
)
around(intersect, (Intersection, object))(
    lambda c1, c2: type(c1)(list(c1)+[c2])
)
around(intersect, (object, Intersection))(
    lambda c1, c2: type(c2)([c1]+list(c2))
)

# Default intersection is a Conjunction
when(intersect, (object, object))(lambda c1, c2: Conjunction([c1,c2]))























class OrElse(Disjunction, tuple):
    """SEQUENCE of or-ed conditions (excluding redundant/implying items)"""

    def __new__(cls, input):
        output = []
        for item in input:
            for old in output[:]:
                if implies(item, old):
                    break
                elif implies(old, item):
                    output.remove(old)
            else:
                output.append(item)
        if not output:
            return False
        elif len(output)==1:
            return output[0]
        return tuple.__new__(cls, output)

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, list(self))


when(disjuncts, (OrElse,))
def sequential_disjuncts(c):
    pre = True
    out = set()
    for cond in c:
        out.update(disjuncts(intersect(pre, cond)))
        pre = intersect(pre, negate(cond))
    return list(out)













from peak.rules.core import when, value
from peak.util.assembler import *
from peak.rules.codegen import *

__all__ = ['precedence', 'associativity', 'decompile',]

def precedence(ob):
    """Get subexpression precedence - 0 is closest-binding precedence
    
    Subexpressions must be parenthesized when their parent expression has
    a lower return value from this function than they do.

    If the parent expression has equal precedence, then grouping is dependent
    upon associativity.
    """
    return 0

def associativity(ob):
    """Subexpression's operator associativity (0 for left, 1 for right)

    If a subexpression and parent expression have equal precedence, parentheses
    are needed unless the subexpression is in this position among the parent
    expression's operands.
    """
    return 0

when(associativity, (Power,) )(value(1))
when(associativity, (IfElse,))(value(2))
when(associativity, (ListOp,))(value(None))   

def needs_parens(parent, child, posn):
    """Does expression `child` need parens if it's at `posn` in `parent`?"""
    return False


def decompile(ob):
    return repr(ob)




prec_ops = (UnaryOp, BinaryOp, ListOp)

for prec, group in enumerate([
    (Tuple, List, Dict, Repr, ListComp, BuildSlice),
    (Getitem, GetSlice, Getattr, Call),
    (Power,), (Plus, Minus, Invert), (Mul, Div, FloorDiv, Mod), (Add, Sub),
    (LeftShift, RightShift), (Bitand,), (Bitxor,), (Bitor,),
    (Compare,), (Not,), (And,), (Or,), (IfElse,),    
]):
    when(precedence, (group,))(value(prec))
    for cls in group:
        if not issubclass(cls, prec_ops):
            prec_ops += (cls,)

when(needs_parens, (prec_ops, prec_ops))
def prec_parens(parent, child, posn):
    pprec = precedence(parent)
    pchild = precedence(child)
    if pprec < pchild: return True
    passoc = associativity(parent)
    return pprec==pchild and passoc is not None and posn != passoc 

when(needs_parens, (Power, (Plus, Minus, Invert)))
def prec_power(next_method, parent, child, posn):
    if posn==1: return False
    return next_method(parent, child, posn)

def decompiled_children(parent, children):
    for posn, child in enumerate(children):
        operand = decompile(child)
        if needs_parens(parent, child, posn):
            operand = '(%s)' % (operand,)
        yield operand
    

when(needs_parens, (Getattr, Const))
def int_parens(parent, child, pons):
    return decompile(child).isdigit()   # parenthesize integers



when(decompile, ((UnaryOp, BinaryOp, GetSlice, IfElse),))
def decompile_fmt(expr):
    return expr.fmt % tuple(decompiled_children(expr, expr[1:]))

when(decompile, (Local,))
def decompile_local(expr):
    return expr[1]

when(decompile, (Getattr,))
def decompile_getattr(expr):
    ignore, left, right = expr
    return '%s.%s' % (list(decompiled_children(expr, [left]))[0], right)

when(decompile, (Pass.__class__,))
def decompile_pass(expr):
    return ''

when(decompile, (Const,))
def decompile_const(expr):
    return decompile(expr.value)

when(decompile, (ListOp,))
def decompile_items(expr):
    return expr.fmt % ', '.join(decompiled_children(expr, expr[1]))

And.separator = ' and '
Or.separator = ' or '

when(decompile, ((And, Or),))
def decompile_sep(expr):
    return expr.separator.join(decompiled_children(expr, expr[1]))
    
when(decompile, (slice,))(
    lambda e: _decompile_slice(e.start, e.stop, e.step)
)
when(decompile, (BuildSlice,))(
    lambda e: _decompile_slice(e.start, e.stop, e.stride)
)



def _decompile_slice(start, stop, step):
    slice = ''
    if start is not None and start is not Pass: slice += decompile(start)
    slice +=':'
    if stop is not None and stop is not Pass: slice += decompile(stop)
    if step is not None and step is not Pass: slice += ':' + decompile(step)
    return slice

when(decompile, (Dict,))
def decompile_dict(expr):
    return '{%s}' % ', '.join(
        [('%s: %s' % tuple(map(decompile, args))) for args in expr[1]]
    )
    
when(decompile, (Call,))
def decompile_dict(expr):
    func = expr[1]
    if needs_parens(expr, func, 0):
        func = '(%s)' % (decompile(func),)
    else:
        func = decompile(func)
    return '%s(%s)' % (func, ', '.join(_decompiled_call_args(expr)))

def _decompiled_call_args(call):
    ignore, ignore, posargs, kwargs, star, dstar, ignore = call
    for arg in posargs:
        yield decompile(arg)
    for kw,arg in kwargs:
        yield '%s=%s' % (kw.value, decompile(arg))
    if star:
        yield '*' + decompile(star)
    if dstar:
        yield '**' + decompile(dstar)



    






















#!/usr/bin/env python
"""Distutils setup file"""
import ez_setup
ez_setup.use_setuptools()
from setuptools import setup
PACKAGE_NAME = "PEAK-Rules"
PACKAGE_VERSION = "0.5a1"
PACKAGES = ['peak', 'peak.rules']

def get_description():
    # Get our long description from the documentation
    f = file('README.txt')
    lines = []
    for line in f:
        if not line.strip():
            break     # skip to first blank line
    for line in f:
        if line.startswith('.. contents::'):
            break     # read to table of contents
        lines.append(line)
    f.close()
    return ''.join(lines)

setup(
    name=PACKAGE_NAME,
    version=PACKAGE_VERSION,
    description="Generic functions and business rules support systems",
    long_description = get_description(),
    install_requires=[
        'BytecodeAssembler>=0.6','DecoratorTools>=1.7dev-r2450',
        'AddOns>=0.6', 'Extremes>=1.1',
    ],
    author="Phillip J. Eby",
    author_email="peak@eby-sarna.com",
    license="ZPL 2.1",
    url="http://pypi.python.org/pypi/PEAK-Rules",
    download_url = "http://peak.telecommunity.com/snapshots/",
    test_suite = 'test_rules', tests_require='Importing',
    packages = PACKAGES,
    namespace_packages = ['peak'],
)
